from mpi4py import MPI

import basix
import dolfinx
import numpy as np
import pytest

from io4dolfinx.readers import create_geometry_function_space


@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.triangle, dolfinx.mesh.CellType.quadrilateral]
)
@pytest.mark.parametrize("value_shape", [(), (1,), (4,)])
@pytest.mark.parametrize("N", [3, 20])
@pytest.mark.parametrize("M", [8, 9])
@pytest.mark.parametrize(
    "ghost_mode", [dolfinx.mesh.GhostMode.shared_facet, dolfinx.mesh.GhostMode.none]
)
def test_dofmap_construction(cell_type, value_shape, N, M, ghost_mode):
    mesh = dolfinx.mesh.create_unit_square(
        MPI.COMM_WORLD, N, M, cell_type=cell_type, ghost_mode=ghost_mode
    )
    el = basix.ufl.element("Lagrange", mesh.basix_cell(), 1)
    if value_shape == ():
        b_el = el
    else:
        b_el = basix.ufl.blocked_element(el, value_shape)

    MPI.COMM_WORLD.barrier()

    V_ref = dolfinx.fem.functionspace(mesh, b_el)

    MPI.COMM_WORLD.barrier()

    bs = int(np.prod(value_shape))
    V = create_geometry_function_space(mesh, bs)
    assert V.dofmap.bs == V_ref.dofmap.bs
    assert V.dofmap.bs == bs
    np.testing.assert_allclose(V.dofmap.list, V_ref.dofmap.list)



"""
Helpers reading/writing data with ADIOS2
"""

from __future__ import annotations

import shutil
from contextlib import contextmanager
from pathlib import Path
from typing import NamedTuple

from mpi4py import MPI

import adios2
import dolfinx.cpp.graph
import dolfinx.graph
import numpy as np
import numpy.typing as npt

from io4dolfinx.utils import compute_local_range, valid_function_types


def resolve_adios_scope(adios2):
    scope = adios2.bindings if hasattr(adios2, "bindings") else adios2
    if not scope.is_built_with_mpi:
        raise ImportError("ADIOS2 must be built with MPI support")
    return scope


adios2 = resolve_adios_scope(adios2)


__all__ = [
    "AdiosFile",
    "ADIOSFile",
    "check_variable_exists",
    "read_array",
    "read_adjacency_list",
    "adios_to_numpy_dtype",
]

adios_to_numpy_dtype = {
    "float": np.float32,
    "double": np.float64,
    "float complex": np.complex64,
    "double complex": np.complex128,
    "uint32_t": np.uint32,
}


class AdiosFile(NamedTuple):
    io: adios2.IO
    file: adios2.Engine


@contextmanager
def ADIOSFile(
    adios: adios2.ADIOS,
    filename: Path | str,
    engine: str,
    mode: adios2.Mode,
    io_name: str,
    comm: MPI.Intracomm | None = None,
):
    io = adios.DeclareIO(io_name)
    io.SetEngine(engine)
    # ADIOS2 sometimes struggles with existing files/folders it should overwrite
    if mode == adios2.Mode.Write:
        filename = Path(filename)
        if filename.exists() and comm is not None and comm.rank == 0:
            if filename.is_dir():
                shutil.rmtree(filename)
            else:
                filename.unlink()
        if comm is not None:
            comm.Barrier()

    file = io.Open(str(filename), mode)
    try:
        yield AdiosFile(io=io, file=file)
    finally:
        file.Close()
        adios.RemoveIO(io_name)


def check_variable_exists(
    adios: adios2.ADIOS,
    filename: Path | str,
    variable: str,
    engine: str,
) -> bool:
    io_name = f"{variable}_reader"

    if not Path(filename).exists():
        return False

    variable_found = False
    with ADIOSFile(
        adios=adios,
        engine=engine,
        filename=filename,
        mode=adios2.Mode.Read,
        io_name=io_name,
    ) as adios_file:
        # Find step that has cell permutation
        for _ in range(adios_file.file.Steps()):
            adios_file.file.BeginStep()
            if variable in adios_file.io.AvailableVariables().keys():
                variable_found = True
                break
            adios_file.file.EndStep()

        # Not sure if this is needed, but just in case
        if variable in adios_file.io.AvailableVariables().keys():
            variable_found = True
    return variable_found


def read_adjacency_list(
    adios: adios2.ADIOS,
    comm: MPI.Intracomm,
    filename: Path | str,
    data_name: str,
    offsets_name: str,
    engine: str,
) -> dolfinx.graph.AdjacencyList:
    """
    Read an adjacency-list from an ADIOS file with given communicator.
    The adjancency list is split in to a flat array (data) and its corresponding offset.

    Args:
        adios: The ADIOS instance
        comm: The MPI communicator used to read the data
        filename: Path to input file
        data_name: Name of variable containing the indices of the adjacencylist
        dofmap_offsets: Name of variable containing offsets of the adjacencylist
        engine: Type of ADIOS engine to use for reading data

    Returns:
        The local part of dofmap from input dofs

    .. note::
        No MPI communication is done during this call
    """

    # Open ADIOS engine
    io_name = f"{data_name=}_reader"

    with ADIOSFile(
        adios=adios,
        engine=engine,
        filename=filename,
        mode=adios2.Mode.Read,
        io_name=io_name,
    ) as adios_file:
        # First find step with dofmap offsets, to be able to read
        # in a full row of the dofmap
        for _ in range(adios_file.file.Steps()):
            adios_file.file.BeginStep()
            if offsets_name in adios_file.io.AvailableVariables().keys():
                break
            adios_file.file.EndStep()
        if offsets_name not in adios_file.io.AvailableVariables().keys():
            raise KeyError(f"Dof offsets not found at '{offsets_name}' in {filename}")

        # Get global shape of dofmap-offset, and read in data with an overlap
        d_offsets = adios_file.io.InquireVariable(offsets_name)
        shape = d_offsets.Shape()
        num_nodes = shape[0] - 1
        local_range = compute_local_range(comm, num_nodes)

        # As the offsets are one longer than the number of cells, we need to read in with an overlap
        if len(shape) == 1:
            d_offsets.SetSelection([[local_range[0]], [local_range[1] + 1 - local_range[0]]])
            in_offsets = np.empty(
                local_range[1] + 1 - local_range[0],
                dtype=d_offsets.Type().strip("_t"),
            )
        else:
            d_offsets.SetSelection(
                [
                    [local_range[0], 0],
                    [local_range[1] + 1 - local_range[0], shape[1]],
                ]
            )
            in_offsets = np.empty(
                (local_range[1] + 1 - local_range[0], shape[1]),
                dtype=d_offsets.Type().strip("_t"),
            )

        adios_file.file.Get(d_offsets, in_offsets, adios2.Mode.Sync)
        in_offsets = in_offsets.squeeze()

        # Assuming dofmap is saved in stame step
        # Get the relevant part of the dofmap
        if data_name not in adios_file.io.AvailableVariables().keys():
            raise KeyError(f"Dofs not found at {data_name} in {filename}")
        cell_dofs = adios_file.io.InquireVariable(data_name)
        if len(shape) == 1:
            cell_dofs.SetSelection([[in_offsets[0]], [in_offsets[-1] - in_offsets[0]]])
            in_dofmap = np.empty(in_offsets[-1] - in_offsets[0], dtype=cell_dofs.Type().strip("_t"))
        else:
            cell_dofs.SetSelection([[in_offsets[0], 0], [in_offsets[-1] - in_offsets[0], shape[1]]])
            in_dofmap = np.empty(
                (in_offsets[-1] - in_offsets[0], shape[1]),
                dtype=cell_dofs.Type().strip("_t"),
            )
            assert shape[1] == 1

        in_dofmap = np.empty(in_offsets[-1] - in_offsets[0], dtype=cell_dofs.Type().strip("_t"))
        adios_file.file.Get(cell_dofs, in_dofmap, adios2.Mode.Sync)
        in_offsets -= in_offsets[0]
        adios_file.file.EndStep()

    # Return local dofmap
    return dolfinx.graph.adjacencylist(in_dofmap, in_offsets.astype(np.int32))


def read_array(
    adios: adios2.ADIOS,
    filename: Path | str,
    array_name: str,
    engine: str,
    comm: MPI.Intracomm,
    time: float = 0.0,
    time_name: str = "",
    legacy: bool = False,
) -> tuple[npt.NDArray[valid_function_types], int]:
    """
    Read an array from file, return the global starting position of the local array

    Args:
        adios: The ADIOS instance
        filename: Path to file to read array from
        array_name: Name of array in file
        engine: Name of engine to use to read file
        comm: MPI communicator used for reading the data
        time_name: Name of time variable for modern checkpoints
        legacy: If True ignore time_name and read the first available step
    Returns:
        Local part of array and its global starting position
    """

    with ADIOSFile(
        adios=adios,
        engine=engine,
        filename=filename,
        mode=adios2.Mode.Read,
        io_name="ArrayReader",
    ) as adios_file:
        # Get time-stamp from first available step
        if legacy:
            for i in range(adios_file.file.Steps()):
                adios_file.file.BeginStep()
                if array_name in adios_file.io.AvailableVariables().keys():
                    break
                adios_file.file.EndStep()
            if array_name not in adios_file.io.AvailableVariables().keys():
                raise KeyError(f"No array found at {array_name}")
        else:
            for i in range(adios_file.file.Steps()):
                adios_file.file.BeginStep()
                if time_name in adios_file.io.AvailableVariables().keys():
                    arr = adios_file.io.InquireVariable(time_name)
                    time_shape = arr.Shape()
                    arr.SetSelection([[0], [time_shape[0]]])
                    times = np.empty(time_shape[0], dtype=adios_to_numpy_dtype[arr.Type()])
                    adios_file.file.Get(arr, times, adios2.Mode.Sync)
                    if times[0] == time:
                        break
                if i == adios_file.file.Steps() - 1:
                    raise KeyError(
                        f"No data associated with {time_name}={time} found in {filename}"
                    )

                adios_file.file.EndStep()

            if time_name not in adios_file.io.AvailableVariables().keys():
                raise KeyError(f"No data associated with {time_name}={time} found in {filename}")

            if array_name not in adios_file.io.AvailableVariables().keys():
                raise KeyError(f"No array found at {time=} for {array_name}")

        arr = adios_file.io.InquireVariable(array_name)
        arr_shape = arr.Shape()
        # TODO: Should we always pick the first element?
        assert len(arr_shape) >= 1
        arr_range = compute_local_range(comm, arr_shape[0])

        if len(arr_shape) == 1:
            arr.SetSelection([[arr_range[0]], [arr_range[1] - arr_range[0]]])
            vals = np.empty(arr_range[1] - arr_range[0], dtype=adios_to_numpy_dtype[arr.Type()])
        else:
            arr.SetSelection([[arr_range[0], 0], [arr_range[1] - arr_range[0], arr_shape[1]]])
            vals = np.empty(
                (arr_range[1] - arr_range[0], arr_shape[1]),
                dtype=adios_to_numpy_dtype[arr.Type()],
            )
            assert arr_shape[1] == 1

        adios_file.file.Get(arr, vals, adios2.Mode.Sync)
        adios_file.file.EndStep()

    return vals.reshape(-1), arr_range[0]



import urllib.request
from enum import Enum

from mpi4py import MPI

import numpy as np
import pytest

import io4dolfinx

netcdf4 = pytest.importorskip("netCDF4")


class DownloadStatus(Enum):
    success = 1
    failed = -1
    no_connection = -2


def download_file_if_not_exists(
    url, filename, comm: MPI.Intracomm = MPI.COMM_WORLD, rank: int = 0
) -> DownloadStatus:
    status = DownloadStatus.failed
    if comm.rank == rank:
        if not filename.exists():
            try:
                urllib.request.urlretrieve(url, filename)
                status = DownloadStatus.success
            except urllib.error.URLError as e:
                if str(e) == "<urlopen error [Errno -3] Temporary failure in name resolution>":
                    status = DownloadStatus.no_connection
                else:
                    status = DownloadStatus.failed
        else:
            status = DownloadStatus.success
    status = comm.bcast(status, root=rank)
    comm.Barrier()
    return status


def test_read_mesh_and_cell_data(tmp_path):
    tmp_path = MPI.COMM_WORLD.bcast(tmp_path, root=0)
    filename = tmp_path / "openmc_master_out_openmc0.e"
    url = "https://github.com/neams-th-coe/cardinal/blob/devel/test/tests/neutronics/feedback/single_level/gold/openmc_master_out_openmc0.e?raw=true"
    status = download_file_if_not_exists(url, filename)
    if status == DownloadStatus.no_connection:
        pytest.skip("No internet connection")
    mesh = io4dolfinx.read_mesh(filename, MPI.COMM_WORLD, backend="exodus")
    io4dolfinx.read_meshtags(filename, mesh, meshtag_name="cell", backend="exodus")
    io4dolfinx.read_meshtags(filename, mesh, meshtag_name="facet", backend="exodus")
    io4dolfinx.read_cell_data(
        filename, name="cell_temperature", mesh=mesh, backend="exodus", time=1.0
    )


def test_read_mesh_point_data(tmp_path):
    tmp_path = MPI.COMM_WORLD.bcast(tmp_path, root=0)

    filename = tmp_path / "openmc_master_out_openmc0.e"
    url = "https://github.com/idaholab/moose/blob/next/test/tests/kernels/2d_diffusion/gold/matdiffusion_out.e?raw=true"
    status = download_file_if_not_exists(url, filename)
    if status == DownloadStatus.no_connection:
        pytest.skip("No internet connection")

    mesh = io4dolfinx.read_mesh(filename, MPI.COMM_WORLD, backend="exodus")
    u = io4dolfinx.read_point_data(filename, name="u", mesh=mesh, backend="exodus", time=1.0)
    assert mesh.topology.index_map(mesh.topology.dim).size_global == 4
    assert mesh.geometry.index_map().size_global == 9
    assert np.isclose(mesh.comm.allreduce(np.max(u.x.array), op=MPI.MAX), 1.1140844375981802)


def test_read_second_order_mesh(tmp_path):
    tmp_path = MPI.COMM_WORLD.bcast(tmp_path, root=0)
    filename = tmp_path / "box-test_out_nek0.e"
    url = "https://github.com/neams-th-coe/cardinal/blob/devel/test/tests/deformation/simple-cube/gold/box-test_out_nek0.e?raw=true"
    status = download_file_if_not_exists(url, filename)
    if status == DownloadStatus.no_connection:
        pytest.skip("No internet connection")

    mesh = io4dolfinx.read_mesh(filename, MPI.COMM_WORLD, backend="exodus", time=5)
    assert mesh.topology.index_map(mesh.topology.dim).size_global == 64
    assert mesh.geometry.index_map().size_global == 1728

    u_x = io4dolfinx.read_point_data(filename, name="disp_x", mesh=mesh, backend="exodus", time=5.0)
    ref_min = -0.7699306351843485
    ref_max = 0.7699306351843485
    assert np.isclose(mesh.comm.allreduce(np.min(u_x.x.array), op=MPI.MIN), ref_min)
    assert np.isclose(mesh.comm.allreduce(np.max(u_x.x.array), op=MPI.MAX), ref_max)



# # Storing mesh partition
# This data is re-ordered when reading in a mesh, as the mesh is partitioned.
# This means that when storing the mesh to disk from DOLFINx, the geometry and
# connectivity arrays are re-ordered.
# If we want to avoid to re-partition the mesh every time you run a simulation
# (on a fixed number of processes), one can store the partitioning of the mesh
# in the checkpoint. This is done by setting the flag `store_partition_info=True`
# when calling {py:func}`io4dolfinx.write_mesh`.

# +
import logging
from pathlib import Path

import ipyparallel as ipp


def write_partitioned_mesh(filename: Path):
    import subprocess

    from mpi4py import MPI

    import dolfinx

    import io4dolfinx

    # Create a simple unit square mesh
    mesh = dolfinx.mesh.create_unit_square(
        MPI.COMM_WORLD,
        10,
        10,
        cell_type=dolfinx.mesh.CellType.quadrilateral,
        ghost_mode=dolfinx.mesh.GhostMode.shared_facet,
    )

    # Write mesh checkpoint
    io4dolfinx.write_mesh(
        filename, mesh, backend="adios2", backend_args={"engine": "BP4"}, store_partition_info=True
    )
    # Inspect checkpoint on rank 0 with `bpls`
    if mesh.comm.rank == 0:
        output = subprocess.run(["bpls", "-a", "-l", filename], capture_output=True)
        print(output.stdout.decode("utf-8"))


# -

# We inspect the partitioned mesh

# +
mesh_file = Path("partitioned_mesh.bp")
n = 3

with ipp.Cluster(engines="mpi", n=n, log_level=logging.ERROR) as cluster:
    query = cluster[:].apply_async(write_partitioned_mesh, mesh_file)
    query.wait()
    assert query.successful(), query.error
    print("".join(query.stdout))
# -

# # Reading a partitioned mesh

# If we try to read the mesh in on a different number of processes, we will get an error.
# We illustrate this below, by first trying to read the mesh using partitioning information,
# which is done by setting the flag `read_from_partition=True` when calling
# {py:func}`io4dolfinx.read_mesh`.


def read_partitioned_mesh(filename: Path, read_from_partition: bool = True):
    from mpi4py import MPI

    import io4dolfinx

    prefix = f"{MPI.COMM_WORLD.rank + 1}/{MPI.COMM_WORLD.size}: "
    try:
        mesh = io4dolfinx.read_mesh(
            filename,
            comm=MPI.COMM_WORLD,
            backend="adios2",
            backend_args={"engine": "BP4"},
            read_from_partition=read_from_partition,
        )
        print(f"{prefix} Mesh: {mesh.name} read successfully with {read_from_partition=}")
    except ValueError as e:
        print(f"{prefix} Caught exception: ", e)


with ipp.Cluster(engines="mpi", n=n + 1, log_level=logging.ERROR) as cluster:
    # Read mesh from file with different number of processes
    query = cluster[:].apply_async(read_partitioned_mesh, mesh_file)
    query.wait()
    assert query.successful(), query.error
    print("".join(query.stdout))

# Read mesh from file with different number of processes (not using partitioning information).
# If we instead turn of `read_from_partition`, we can read the mesh on a
# different number of processes.

with ipp.Cluster(engines="mpi", n=n + 1, log_level=logging.ERROR) as cluster:
    query = cluster[:].apply_async(read_partitioned_mesh, mesh_file, False)
    query.wait()
    assert query.successful(), query.error
    print("".join(query.stdout))

# Read mesh from file with same number of processes as was written,
# re-using partitioning information.

with ipp.Cluster(engines="mpi", n=n, log_level=logging.ERROR) as cluster:
    query = cluster[:].apply_async(read_partitioned_mesh, mesh_file, True)
    query.wait()
    assert query.successful(), query.error
    print("".join(query.stdout))



"""
Exodus interface to io4dolfinx.
The Exodus2 format is described in:
https://src.fedoraproject.org/repo/pkgs/exodusii/922137.pdf/a45d67f4a1a8762bcf66af2ec6eb35f9/922137.pdf
Further documentation from CUBIT on node numbering can be found at:
https://coreform.com/cubit_help/appendix/element_numbering.htm

SPDX License identifier: MIT

Copyright: Jørgen S. Dokken, Henrik N.T. Finsberg, Remi Delaporte-Mathurin,
           and Simula Research Laboratory
"""

from pathlib import Path
from typing import Any, Literal, cast

from mpi4py import MPI

import basix.ufl
import dolfinx
import netCDF4
import numpy as np
import numpy.typing as npt

from ...structures import ArrayData, FunctionData, MeshData, MeshTagsData, ReadMeshData
from ...utils import check_file_exists
from .. import FileMode, ReadMode

_interval_to_vertex_map = {0: [0, 1]}
# Based on: https://src.fedoraproject.org/repo/pkgs/exodusii/922137.pdf/a45d67f4a1a8762bcf66af2ec6eb35f9/922137.pdf
_tetra_facet_to_vertex_map = {0: [0, 1, 3], 1: [1, 2, 3], 2: [0, 2, 3], 3: [0, 1, 2]}
# https://coreform.com/cubit_help/appendix/element_numbering.htm
# Note that triangular side-sets goes from 2 to 4 (with 0 base index)
_triangle_to_vertex_map = {2: [0, 1], 3: [1, 2], 4: [2, 0]}
_quad_to_vertex_map = {0: [0, 1], 1: [1, 2], 2: [2, 3], 3: [3, 0]}
_hex_to_vertex_map = {
    0: [0, 1, 4, 5],
    1: [1, 2, 5, 6],
    2: [2, 3, 6, 7],
    3: [0, 3, 4, 7],
    4: [0, 1, 2, 3],
    5: [4, 5, 6, 7],
}

_side_set_to_vertex_map = {
    "interval": _interval_to_vertex_map,
    "quadrilateral": _quad_to_vertex_map,
    "triangle": _triangle_to_vertex_map,
    "tetrahedron": _tetra_facet_to_vertex_map,
    "hexahedron": _hex_to_vertex_map,
}


_exodus_to_string = {
    "EDGE2": ("interval", 1),
    "TRI3": ("triangle", 1),
    "QUAD": ("quadrilateral", 1),
    "QUAD4": ("quadrilateral", 1),
    "TETRA": ("tetrahedron", 1),
    "HEX": ("hexahedron", 1),
    "HEX8": ("hexahedron", 1),
    "BEAM2": ("interval", 1),
    "HEX27": ("hexahedron", 2),
}

read_mode = ReadMode.serial


def _get_cell_type(connectivity: netCDF4.Variable) -> tuple[dolfinx.mesh.CellType, int]:
    cell_type, degree = _exodus_to_string[connectivity.elem_type]
    return dolfinx.mesh.to_type(cell_type), degree


def _compute_tdim(connectivity: netCDF4.Variable) -> int:
    d_ct, _ = _get_cell_type(connectivity)
    return dolfinx.mesh.cell_dim(d_ct)


def get_default_backend_args(arguments: dict[str, Any] | None) -> dict[str, Any]:
    args = arguments or {}
    return args


def convert_file_mode(mode: FileMode) -> str:
    match mode:
        case FileMode.append:
            return "a"
        case FileMode.read:
            return "r"
        case FileMode.write:
            return "w"
        case _:
            raise NotImplementedError(f"File mode {mode} not implemented")


def write_attributes(
    filename: Path | str,
    comm: MPI.Intracomm,
    name: str,
    attributes: dict[str, np.ndarray],
    backend_args: dict[str, Any] | None = None,
):
    """Write attributes to file using H5PY.

    Args:
        filename: Path to file to write to
        comm: MPI communicator used in storage
        name: Name of the attributes
        attributes: Dictionary of attributes to write to file
        backend_args: Arguments to backend
    """
    raise NotImplementedError("The Exodus backend cannot write attributes.")


def read_attributes(
    filename: Path | str,
    comm: MPI.Intracomm,
    name: str,
    backend_args: dict[str, Any] | None = None,
) -> dict[str, Any]:
    """Read attributes from file using H5PY.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the attributes
        backend_args: Arguments to backend
    Returns:
        The attributes
    """
    raise NotImplementedError("The Exodus backend cannot read attributes.")


def snapshot_checkpoint(
    filename: Path | str,
    mode: FileMode,
    u: dolfinx.fem.Function,
    backend_args: dict[str, Any] | None,
):
    """Create a snapshot checkpoint of a dolfinx function.

    Args:
        filename: Path to file to read from
        mode: File-mode to store the function
        u: dolfinx function to create a snapshot checkpoint for
        backend_args: Arguments to backend
    """
    raise NotImplementedError("The EXODUS backend cannot make checkpoints.")


def read_hdf5_array(
    comm: MPI.Intracomm,
    filename: Path | str,
    group: str,
    backend_args: dict[str, Any] | None,
) -> tuple[np.ndarray, int]:
    """Read an array from an HDF5 file.

    Args:
        comm: MPI communicator used in storage
        filename: Path to file to read from
        group: Group in HDF5 file where array is stored
        backend_args: Arguments to backend

    Returns:
        Tuple containing:
            - Numpy array read from file
            - Global starting point on the process.
                Process 0 has [0, M), process 1 [M, N), process 2 [N, O) etc.
    """
    raise NotImplementedError("The EXODUS backend cannot read HDF5 arrays")


def read_timestamps(
    filename: Path | str,
    comm: MPI.Intracomm,
    function_name: str,
    backend_args: dict[str, Any] | None = None,
) -> npt.NDArray[np.float64 | str]:  # type: ignore[type-var]
    """Read time-stamps from a checkpoint file.

    Args:
        comm: MPI communicator
        filename: Path to file
        comm: MPI communicator
        function_name: Name of the function to read time-stamps for
        backend_args: Arguments for backend, for instance file type.

    Returns:
        The time-stamps
    """
    raise NotImplementedError("The Exodus backend cannot read timestamps.")


def write_mesh(
    filename: Path | str,
    comm: MPI.Intracomm,
    mesh: MeshData,
    backend_args: dict[str, Any] | None = None,
    mode: FileMode = FileMode.write,
    time: float = 0.0,
):
    """Write a mesh to file using H5PY

    Args:
        filename: Path to file to write to.
        mesh: Internal data structure for the mesh data to save to file
        comm: MPI communicator used in storage
        backend_args: Arguments to backend
        mode: Mode to use (write or append)
        time: Time stamp
    """
    raise NotImplementedError("The Exodus backend cannot write meshes.")


def _read_mesh_geometry(infile: netCDF4.Dataset) -> tuple[int, npt.NDArray[np.floating]]:
    # use page 171 of manual to extract data
    num_nodes = infile.dimensions["num_nodes"].size
    gdim = infile.dimensions["num_dim"].size

    # Get coordinates of mesh
    coord_var = infile.variables.get("coord")
    if coord_var is None:
        coordinates = np.zeros((num_nodes, gdim), dtype=np.float64)
        for i, coord in enumerate(["x", "y", "z"]):
            coord_i = infile.variables.get(f"coord{coord}")
            if coord_i is not None:
                coordinates[: coord_i.size, i] = coord_i[:]
    else:
        coordinates = np.asarray(coord_var)
    return gdim, coordinates


def _get_entity_blocks(
    infile: netCDF4.Dataset, search_type: Literal["cell", "facet"]
) -> tuple[int, list[netCDF4.Variable]]:
    # use page 171 of manual to extract data
    num_blocks = infile.dimensions["num_el_blk"].size

    # Get element connectivity
    all_connectivity_variables = [infile.variables[f"connect{i + 1}"] for i in range(num_blocks)]

    # Compute max topological dimension in mesh and find the correct
    max_tdim = _compute_tdim(max(all_connectivity_variables, key=_compute_tdim))

    # Extract only the connectivity blocks that we need
    if search_type == "cell":
        search_dim = max_tdim
    elif search_type == "facet":
        search_dim = max_tdim - 1
    else:
        raise RuntimeError(f"Unknown entity type: {search_type}")
    return search_dim, list(
        filter(lambda el: _compute_tdim(el) == search_dim, all_connectivity_variables)
    )


def _extract_connectivity_data(
    entity_blocks: list[netCDF4.Variable],
) -> tuple[list[npt.NDArray[np.int64]], tuple[dolfinx.mesh.CellType, int], list[int]]:
    connectivity_arrays = []
    cell_types = []
    entity_block_index = []
    for entity_block in entity_blocks:
        connectivity_arrays.append(entity_block[:] - 1)
        cell_types.append(_get_cell_type(entity_block))
        entity_block_index.append(int(entity_block.name.removeprefix("connect")) - 1)
    for cell in cell_types:
        assert cell_types[0] == cell, "Mixed cell types not supported"
    cell_type = cell_types[0]
    return connectivity_arrays, cell_type, entity_block_index


def read_mesh_data(
    filename: Path | str,
    comm: MPI.Intracomm,
    time: str | float | None = 0.0,
    read_from_partition: bool = False,
    backend_args: dict[str, Any] | None = None,
) -> ReadMeshData:
    """Read mesh data from EXODUS based checkpoint files.

    Args:
        filename: Path to input file
        comm: The MPI communciator to distribute the mesh over
        time: Time stamp associated with mesh
        read_from_partition: Read mesh with partition from file
        backend_args: Arguments to backend
    Returns:
        The mesh topology, geometry, UFL domain and partition function
    """
    check_file_exists(filename)
    with netCDF4.Dataset(filename, "r") as infile:
        if comm.rank == 0:
            gdim, coordinates = _read_mesh_geometry(infile)

            _tdim, entity_blocks = _get_entity_blocks(infile, "cell")
            if len(entity_blocks) > 0:
                # Extract markers directly from entity-blocks
                connectivity_arrays, (cell_type, degree), _entity_block_index = (
                    _extract_connectivity_data(entity_blocks)
                )

                cells = np.vstack(connectivity_arrays)
                if isinstance(cells, np.ma.MaskedArray):
                    cells = cells.filled()
            else:
                raise ValueError(f"No blocks found in {filename}")

            if degree == 1:
                perm = dolfinx.cpp.io.perm_vtk(cell_type, cells.shape[1])
            elif cell_type == dolfinx.mesh.CellType.hexahedron and degree == 2:
                # Ordering from Fig 4.14 of: https://sandialabs.github.io/seacas-docs/exodusII-new.pdf
                dolfinx_to_exodus = np.array(
                    [
                        0,
                        1,
                        5,
                        4,
                        2,
                        3,
                        7,
                        6,
                        8,
                        12,
                        16,
                        10,
                        9,
                        11,
                        18,
                        17,
                        13,
                        15,
                        19,
                        14,
                        26,
                        21,
                        24,
                        22,
                        23,
                        20,
                        25,
                    ]
                )
                perm = np.argsort(dolfinx_to_exodus)
            else:
                raise NotImplementedError(
                    "Reading Exodus2 mesh with {cell_type} of order {degree} is not supported."
                )

            cells = cells[:, perm]
            cell_type, gdim, xtype, degree, num_dofs_per_cell = comm.bcast(
                (cell_type, gdim, np.dtype(coordinates.dtype).name, degree, cells.shape[1]), root=0
            )

        else:
            cell_type, gdim, xtype, degree, num_dofs_per_cell = comm.bcast(
                (None, None, None, None), root=0
            )
            coordinates = np.zeros((0, gdim), dtype=xtype)
            cells = np.zeros((0, num_dofs_per_cell), dtype=np.int64)
    return ReadMeshData(
        cells=cells,
        cell_type=dolfinx.mesh.to_string(cell_type),
        x=coordinates,
        lvar=int(basix.LagrangeVariant.equispaced),
        degree=degree,
        partition_graph=None,
    )


def write_meshtags(
    filename: str | Path,
    comm: MPI.Intracomm,
    data: MeshTagsData,
    backend_args: dict[str, Any] | None = None,
):
    """Write mesh tags to file.

    Args:
        filename: Path to file to write to
        comm: MPI communicator used in storage
        data: Internal data structure for the mesh tags to save to file
        backend_args: Arguments to backend
    """
    raise NotImplementedError("The Exodus backend cannot write meshtags.")


def read_meshtags_data(
    filename: str | Path,
    comm: MPI.Intracomm,
    name: str,
    backend_args: dict[str, Any] | None = None,
) -> MeshTagsData:
    """Read mesh tags from file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the mesh tags to read
        backend_args: Arguments to backend.

    Returns:
        Internal data structure for the mesh tags read from file
    """
    if comm.rank == 0:
        with netCDF4.Dataset(filename, "r") as infile:
            # Compute max topological dimension in mesh and find the correct
            if name == "cell" or name == "facet":
                search_dim, entity_blocks = _get_entity_blocks(
                    infile, cast(Literal["cell", "facet"], name)
                )
            else:
                raise RuntimeError("Expected name='cell' or 'facet' got {name}")

            if len(entity_blocks) > 0:
                # Extract markers directly from entity-blocks
                connectivity_arrays, (cell_type, degree), entity_block_index = (
                    _extract_connectivity_data(entity_blocks)
                )
                marked_entities = np.vstack(connectivity_arrays)
                entity_values = np.zeros(marked_entities.shape[0], dtype=np.int64)
                if "eb_prop1" in infile.variables.keys():
                    block_values = infile.variables["eb_prop1"][:]

                    # First check if entities are in eb_prop1
                    insert_offset = np.zeros(len(connectivity_arrays) + 1, dtype=np.int64)
                    insert_offset[1:] = np.cumsum([c_arr.shape[0] for c_arr in connectivity_arrays])
                    for i, index in enumerate(entity_block_index):
                        entity_values[insert_offset[i] : insert_offset[i + 1]] = block_values[index]
                else:
                    num_dofs_per_cell = basix.ufl.element(
                        "Lagrange", dolfinx.mesh.to_string(cell_type), degree
                    ).dim
                    assert num_dofs_per_cell == marked_entities.shape[1]
                    marked_entities = np.zeros((0, marked_entities.shape[1]), dtype=np.int64)
                    entity_values = np.zeros(0, dtype=np.int64)
            elif name == "facet" and "ss_prop1" in infile.variables.keys():
                # If we haven't found the cell type as a block, we should be extracting facets
                # (from side-sets), then we need the parent cell
                tdim, entity_blocks = _get_entity_blocks(infile, "cell")
                cell_types = []
                for entity_block in entity_blocks:
                    cell_types.append(_get_cell_type(entity_block))
                for cell in cell_types:
                    assert cell_types[0] == cell, "Mixed cell types not supported"
                cell_type, degree = cell_types[0]
                local_facet_index = _side_set_to_vertex_map[dolfinx.mesh.to_string(cell_type)]
                if "num_side_sets" not in infile.dimensions:
                    facet_type = dolfinx.cpp.mesh.cell_entity_type(cell_type, tdim - 1, 0)
                    num_dofs_per_cell = basix.ufl.element(
                        "Lagrange", dolfinx.mesh.to_string(facet_type), degree
                    ).dim
                    marked_entities = np.zeros((0, num_dofs_per_cell), dtype=np.int64)
                    entity_values = np.zeros(0, dtype=np.int64)
                else:
                    # Extract facet values
                    local_facet_index = _side_set_to_vertex_map[dolfinx.mesh.to_string(cell_type)]
                    num_facet_sets = infile.dimensions["num_side_sets"].size
                    values = infile.variables["ss_prop1"]
                    # Extract all cell blocks to get correct look-up
                    connectivity_arrays = []
                    for entity_block in entity_blocks:
                        connectivity_arrays.append(entity_block[:] - 1)
                    connectivity_array = np.vstack(connectivity_arrays)

                    # Loop through all side sets to extract the correct connectivity
                    facet_indices = []
                    facet_values = []
                    for i in range(num_facet_sets):
                        value = values[i].reshape(-1)
                        elements = infile.variables[f"elem_ss{i + 1}"]
                        local_facets = infile.variables[f"side_ss{i + 1}"]
                        for element, index in zip(elements, local_facets):
                            facet_indices.append(
                                connectivity_array[element - 1, local_facet_index[index - 1]]
                            )
                            facet_values.append(value.data.tolist())
                        marked_entities = np.vstack(facet_indices)
                        entity_values = np.array(facet_values, dtype=np.int64).flatten()
            else:
                # If we found no blocks (for instance for facets, we search through the cells)
                tdim, entity_blocks = _get_entity_blocks(infile, "cell")
                search_dim = tdim - 1 if name == "facet" else tdim
                cell_type, degree = _get_cell_type(entity_blocks[0])
                facet_type = dolfinx.cpp.mesh.cell_entity_type(cell_type, search_dim, 0)
                num_dofs_per_cell = basix.ufl.element(
                    "Lagrange", dolfinx.mesh.to_string(facet_type), degree
                ).dim
                # If we cannot find any information about the blocks we send nothing
                marked_entities = np.zeros((0, num_dofs_per_cell), dtype=np.int64)
                entity_values = np.zeros(0, dtype=np.int64)
        # Broadcast information read by this process to other processes
        (dim, _, _) = comm.bcast(
            (search_dim, marked_entities.shape[1], np.dtype(entity_values.dtype).name),
            root=0,
        )

    else:
        # Other process gets info from process that read the file
        dim, num_dofs_per_cell, vtype = comm.bcast((None, None, None), root=0)
        marked_entities = np.zeros((0, num_dofs_per_cell), dtype=np.int64)
        entity_values = np.zeros(0, dtype=vtype)
    return MeshTagsData(name=name, values=entity_values, indices=marked_entities, dim=dim)


def read_dofmap(
    filename: str | Path,
    comm: MPI.Intracomm,
    name: str,
    backend_args: dict[str, Any] | None,
) -> dolfinx.graph.AdjacencyList:
    """Read the dofmap of a function with a given name.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the function to read the dofmap for
        backend_args: Arguments to backend

    Returns:
        Dofmap as an {py:class}`dolfinx.graph.AdjacencyList`
    """
    raise NotImplementedError("The Exodus backend cannot read dofmap.")


def read_dofs(
    filename: str | Path,
    comm: MPI.Intracomm,
    name: str,
    time: float,
    backend_args: dict[str, Any] | None,
) -> tuple[npt.NDArray[np.float32 | np.float64 | np.complex64 | np.complex128], int]:
    """Read the dofs (values) of a function with a given name from a given timestep.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the function to read the dofs for
        time: Time stamp associated with the function to read
        backend_args: Arguments to backend

    Returns:
        Contiguous sequence of degrees of freedom (with respect to input data)
        and the global starting point on the process.
        Process 0 has [0, M), process 1 [M, N), process 2 [N, O) etc.
    """
    raise NotImplementedError("The Exodus backend cannot read dofs.")


def read_cell_perms(
    comm: MPI.Intracomm, filename: Path | str, backend_args: dict[str, Any] | None
) -> npt.NDArray[np.uint32]:
    """
    Read cell permutation from file with given communicator,
    Split in continuous chunks based on number of cells in the input data.

    Args:
        comm: MPI communicator used in storage
        filename: Path to file to read from
        backend_args: Arguments to backend

    Returns:
        Contiguous sequence of permutations (with respect to input data)
        Process 0 has [0, M), process 1 [M, N), process 2 [N, O) etc.
    """

    raise NotImplementedError("The Exodus backend cannot read cell perms.")


def write_function(
    filename: str | Path,
    comm: MPI.Intracomm,
    u: FunctionData,
    time: float,
    mode: FileMode,
    backend_args: dict[str, Any] | None = None,
):
    """Write a function to file.

    Args:
        filename: Path to file to write to
        comm: MPI communicator used in storage
        u: Internal data structure for the function data to save to file
        time: Time stamp associated with function
        mode: File-mode to store the function
        backend_args: Arguments to backend
    """

    raise NotImplementedError("The Exodus backend cannot write function.")


def read_legacy_mesh(
    filename: Path | str, comm: MPI.Intracomm, group: str
) -> tuple[npt.NDArray[np.int64], npt.NDArray[np.floating], str | None]:
    """Read in the mesh topology, geometry and (optionally) cell type from a
    legacy DOLFIN HDF5-file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        group: Group in HDF5 file where mesh is stored

    Returns:
        Tuple containing:
            - Topology as a (num_cells, num_vertices_per_cell) array of global vertex indices
            - Geometry as a (num_vertices, geometric_dimension) array of vertex coordinates
            - Cell type as a string (e.g. "tetrahedron") or None if not found
    """
    raise NotImplementedError("The Exodus backend cannot read legacy mesh.")


def read_point_data(
    filename: Path | str,
    name: str,
    comm: MPI.Intracomm,
    time: float | str | None,
    backend_args: dict[str, Any] | None,
) -> tuple[np.ndarray, int]:
    """Read data from the nodes of a mesh.

    Args:
        filename: Path to file
        name: Name of point data
        comm: Communicator to launch IO on.
        time: The time stamp
        backend_args: The backend arguments

    Returns:
       Data local to process (contiguous, no mpi comm) and local start range
    """
    num_components = 1  # Default assumption, overriden by data read in having multiple components
    if comm.rank == 0:
        with netCDF4.Dataset(filename, "r") as infile:
            raw_names = infile.variables["name_nod_var"][:].data
            node_names = netCDF4.chartostring(raw_names)
            if name not in node_names:
                raise ValueError(
                    f"Point data with name {name} not found in file.",
                    f"Available variables: {node_names}",
                )
            index = np.flatnonzero(name == node_names)[0] + 1

            temporal_dataset = infile.variables[f"vals_nod_var{index}"]
            time_steps = infile.variables["time_whole"][:].data
            if time is None:
                time_idx = time_steps[0]
            else:
                time_indices = np.flatnonzero(np.isclose(time_steps, time))
                if len(time_indices) == 0:
                    raise ValueError(
                        f"Could not find {name}(t={time}), available time steps are {time_steps}"
                    )
                time_idx = time_indices[0]

            dataset = temporal_dataset[time_idx]
            if len(dataset.shape) == 1:
                dataset = dataset.reshape(-1, num_components)
            else:
                num_components = dataset.shape[1]
    # Broadcast num components to all other ranks
    num_components = comm.bcast(num_components, root=0)
    # Zero data on all other processes
    if comm.rank != 0:
        dataset = np.zeros((0, num_components), dtype=np.float64)
    return dataset, 0


def read_cell_data(
    filename: Path | str,
    name: str,
    comm: MPI.Intracomm,
    time: str | float | None,
    backend_args: dict[str, Any] | None,
) -> tuple[npt.NDArray[np.int64], np.ndarray]:
    """Read data from the cells of a mesh.

    Args:
        filename: Path to file
        name: Name of point data
        comm: Communicator to launch IO on.
        time: The time stamp
        backend_args: The backend arguments
    Returns:
        A tuple (topology, dofs) where topology contains the
        vertex indices of the cells, dofs the degrees of
        freedom within that cell.
    """
    num_components = 1  # Default assumption, overriden by data read in having multiple components
    if comm.rank == 0:
        with netCDF4.Dataset(filename, "r") as infile:
            raw_names = infile.variables["name_elem_var"][:].data
            num_blocks = infile.dimensions["num_el_blk"].size

            node_names = netCDF4.chartostring(raw_names)
            if name not in node_names:
                raise ValueError(
                    f"Cell data with name {name} not found in file.",
                    f"Available variables: {node_names}",
                )
            index = np.flatnonzero(name == node_names)[0] + 1

            entity_blocks = [
                infile.variables[f"vals_elem_var{index}eb{i + 1}"] for i in range(num_blocks)
            ]
            time_steps = infile.variables["time_whole"][:].data
            if time is None:
                time_idx = time_steps[0]
            else:
                time_indices = np.flatnonzero(np.isclose(time_steps, time))
                if len(time_indices) == 0:
                    raise ValueError(
                        f"Could not find {name}(t={time}), available time steps are {time_steps}"
                    )
                time_idx = time_indices[0]

            if len(entity_blocks) > 0:
                datasets = []
                for entity_block in entity_blocks:
                    datasets.append(entity_block[time_idx])

            dataset = np.hstack(datasets)

            if len(dataset.shape) == 1:
                dataset = dataset.reshape(-1, num_components)
            else:
                num_components = dataset.shape[1]
    # Broadcast num components to all other ranks
    num_components = comm.bcast(num_components, root=0)

    # Zero data on all other processes
    if comm.rank != 0:
        dataset = np.zeros((0, num_components), dtype=np.float64)
    _time = float(time) if time is not None else None

    topology = read_mesh_data(filename, comm, _time, False, backend_args=None).cells
    return topology, dataset


def read_function_names(
    filename: Path | str, comm: MPI.Intracomm, backend_args: dict[str, Any] | None
) -> list[str]:
    """Read all function names from a file.

    Args:
        filename: Path to file
        comm: MPI communicator to launch IO on.
        backend_args: Arguments to backend

    Returns:
        A list of function names.
    """
    with netCDF4.Dataset(filename, "r") as infile:
        function_names: list[str] = []
        for key in ["name_elem_var", "name_nod_var"]:
            raw_names = infile.variables[key][:].data
            decoded_names = netCDF4.chartostring(raw_names)
            function_names.extend(decoded_names)
    return function_names


def write_data(
    filename: Path | str,
    array_data: ArrayData,
    comm: MPI.Intracomm,
    time: str | float | None,
    mode: FileMode,
    backend_args: dict[str, Any] | None,
):
    """Write a 2D-array to file (distributed across proceses with MPI).

    Args:
        filename: Path to file
        array_data: Data to write to file
        comm: MPI communicator to open the file with
        time: Time-stamp for data.
        mode: Append or write
        backend_args: The backend arguments
    """
    raise NotImplementedError("Exodus has not implemented this yet")



import itertools
from pathlib import Path

from mpi4py import MPI

import basix
import basix.ufl
import dolfinx
import numpy as np
import pytest

import io4dolfinx

dtypes = [np.float64, np.float32]  # Mesh geometry dtypes
write_comm = [MPI.COMM_SELF, MPI.COMM_WORLD]  # Communicators for creating mesh

two_dimensional_cell_types = [
    dolfinx.mesh.CellType.triangle,
    dolfinx.mesh.CellType.quadrilateral,
]
three_dimensional_cell_types = [
    dolfinx.mesh.CellType.tetrahedron,
    dolfinx.mesh.CellType.hexahedron,
]

two_dim_combinations = itertools.product(dtypes, two_dimensional_cell_types, write_comm)
three_dim_combinations = itertools.product(dtypes, three_dimensional_cell_types, write_comm)


@pytest.fixture(params=two_dim_combinations, scope="module")
def mesh_2D(request):
    dtype, cell_type, write_comm = request.param
    mesh = dolfinx.mesh.create_unit_square(write_comm, 10, 10, cell_type=cell_type, dtype=dtype)
    return mesh


@pytest.fixture(params=three_dim_combinations, scope="module")
def mesh_3D(request):
    dtype, cell_type, write_comm = request.param
    M = 5
    mesh = dolfinx.mesh.create_unit_cube(write_comm, M, M, M, cell_type=cell_type, dtype=dtype)
    return mesh


@pytest.mark.parametrize("is_complex", [True, False])
@pytest.mark.parametrize("family", ["Lagrange", "DG"])
@pytest.mark.parametrize("degree", [1, 4])
@pytest.mark.parametrize("read_comm", [MPI.COMM_SELF, MPI.COMM_WORLD])
def test_read_write_P_2D(
    read_comm,
    family,
    degree,
    is_complex,
    mesh_2D,
    get_dtype,
    write_function,
    read_function,
    backend,
):
    mesh = mesh_2D
    f_dtype = get_dtype(mesh.geometry.x.dtype, is_complex)

    el = basix.ufl.element(
        family,
        mesh.basix_cell(),
        degree,
        basix.LagrangeVariant.gll_warped,
        shape=(mesh.geometry.dim,),
        dtype=mesh.geometry.x.dtype,
    )

    def f(x):
        values = np.empty((2, x.shape[1]), dtype=f_dtype)
        values[0] = np.full(x.shape[1], np.pi) + x[0]
        values[1] = x[0]
        if is_complex:
            values[0] += 1j * x[1]
            values[1] -= 3j * x[1]
        return values

    hash = write_function(mesh, el, f, f_dtype, backend=backend)
    MPI.COMM_WORLD.Barrier()
    read_function(read_comm, el, f, hash, f_dtype, backend=backend)


@pytest.mark.parametrize("is_complex", [True, False])
@pytest.mark.parametrize("family", ["Lagrange", "DG"])
@pytest.mark.parametrize("degree", [1, 4])
@pytest.mark.parametrize("read_comm", [MPI.COMM_SELF, MPI.COMM_WORLD])
def test_read_write_P_3D(
    read_comm,
    family,
    degree,
    is_complex,
    mesh_3D,
    get_dtype,
    write_function,
    read_function,
    backend,
):
    mesh = mesh_3D
    f_dtype = get_dtype(mesh.geometry.x.dtype, is_complex)
    el = basix.ufl.element(
        family,
        mesh.basix_cell(),
        degree,
        basix.LagrangeVariant.gll_warped,
        shape=(mesh.geometry.dim,),
        dtype=mesh.geometry.x.dtype,
    )

    def f(x):
        values = np.empty((3, x.shape[1]), dtype=f_dtype)
        values[0] = np.pi + x[0]
        values[1] = x[1] + 2 * x[0]
        values[2] = np.cos(x[2])
        if is_complex:
            values[0] -= 2j * x[2]
            values[2] += 1j * x[1]
        return values

    hash = write_function(mesh, el, f, f_dtype, backend=backend)

    MPI.COMM_WORLD.Barrier()
    read_function(read_comm, el, f, hash, f_dtype, backend=backend)


@pytest.mark.parametrize("is_complex", [True, False])
@pytest.mark.parametrize("family", ["Lagrange", "DG"])
@pytest.mark.parametrize("degree", [1, 4])
@pytest.mark.parametrize("read_comm", [MPI.COMM_SELF, MPI.COMM_WORLD])
def test_read_write_P_2D_time(
    read_comm,
    family,
    degree,
    is_complex,
    mesh_2D,
    get_dtype,
    write_function_time_dep,
    read_function_time_dep,
    backend,
):
    mesh = mesh_2D
    f_dtype = get_dtype(mesh.geometry.x.dtype, is_complex)

    el = basix.ufl.element(
        family,
        mesh.basix_cell(),
        degree,
        basix.LagrangeVariant.gll_warped,
        shape=(mesh.geometry.dim,),
        dtype=mesh.geometry.x.dtype,
    )

    def f0(x):
        values = np.empty((2, x.shape[1]), dtype=f_dtype)
        values[0] = np.full(x.shape[1], np.pi) + x[0]
        values[1] = x[0]
        if is_complex:
            values[0] += x[1] * 1j
            values[1] -= 3j * x[1]
        return values

    def f1(x):
        values = np.empty((2, x.shape[1]), dtype=f_dtype)
        values[0] = 2 * np.full(x.shape[1], np.pi) + x[0]
        values[1] = -x[0] + 2 * x[1]
        if is_complex:
            values[0] += x[1] * 1j
            values[1] += 3j * x[1]
        return values

    t0 = 0.8
    t1 = 0.6
    hash = write_function_time_dep(mesh, el, f0, f1, t0, t1, f_dtype, backend)
    MPI.COMM_WORLD.Barrier()
    read_function_time_dep(read_comm, el, f0, f1, t0, t1, hash, f_dtype, backend)


@pytest.mark.parametrize("is_complex", [True, False])
@pytest.mark.parametrize("family", ["Lagrange", "DG"])
@pytest.mark.parametrize("degree", [1, 4])
@pytest.mark.parametrize("read_comm", [MPI.COMM_SELF, MPI.COMM_WORLD])
def test_read_write_P_3D_time(
    read_comm,
    family,
    degree,
    is_complex,
    mesh_3D,
    get_dtype,
    write_function_time_dep,
    read_function_time_dep,
    backend,
):
    mesh = mesh_3D
    f_dtype = get_dtype(mesh.geometry.x.dtype, is_complex)
    el = basix.ufl.element(
        family,
        mesh.basix_cell(),
        degree,
        basix.LagrangeVariant.gll_warped,
        shape=(mesh.geometry.dim,),
        dtype=mesh.geometry.x.dtype,
    )

    def f(x):
        values = np.empty((3, x.shape[1]), dtype=f_dtype)
        values[0] = np.pi + x[0]
        values[1] = x[1] + 2 * x[0]
        values[2] = np.cos(x[2])
        if is_complex:
            values[0] += 2j * x[2]
            values[2] += 5j * x[1]
        return values

    def g(x):
        values = np.empty((3, x.shape[1]), dtype=f_dtype)
        values[0] = x[0]
        values[1] = 2 * x[0]
        values[2] = x[0]
        if is_complex:
            values[0] += np.pi * 2j * x[2]
            values[1] += 1j * x[2]
            values[2] += 1j * np.cos(x[1])
        return values

    t0 = 0.1
    t1 = 1.3
    hash = write_function_time_dep(mesh, el, g, f, t0, t1, f_dtype, backend)
    MPI.COMM_WORLD.Barrier()
    read_function_time_dep(read_comm, el, g, f, t0, t1, hash, f_dtype, backend)


@pytest.mark.parametrize(
    "func, args",
    [
        (io4dolfinx.read_attributes, ("nonexisting_file", MPI.COMM_WORLD, "")),
        (io4dolfinx.read_timestamps, ("nonexisting_file", MPI.COMM_WORLD, "")),
        (io4dolfinx.read_meshtags, ("nonexisting_file", MPI.COMM_WORLD, None, "")),
        (io4dolfinx.read_function, ("nonexisting_file", None)),
        (io4dolfinx.read_mesh, ("nonexisting_file", MPI.COMM_WORLD)),
    ],
)
def test_read_nonexisting_file_raises_FileNotFoundError(func, args, backend):
    if backend == "adios":
        suffix = ".bp"
    else:
        suffix = ".h5"
    file = Path(args[0]).with_suffix(suffix)
    with pytest.raises(FileNotFoundError):
        func(file, *args[1:], backend=backend)


def test_read_function_with_invalid_name_raises_KeyError(tmp_path, backend):
    if backend == "adios":
        suffix = ".bp"
    else:
        suffix = ".h5"
    comm = MPI.COMM_WORLD
    f_path = comm.bcast(tmp_path, root=0)
    filename = (f_path / "func").with_suffix(suffix)
    mesh = dolfinx.mesh.create_unit_square(comm, 10, 10, cell_type=dolfinx.mesh.CellType.triangle)
    V = dolfinx.fem.functionspace(mesh, ("P", 1))
    u = dolfinx.fem.Function(V)
    io4dolfinx.write_function(filename, u, time=0, name="some_name", backend=backend)
    io4dolfinx.write_function(filename, u, time=0, name="some_other_name", backend=backend)
    # variables = set(sorted(["some_name", "some_other_name"]))
    with pytest.raises(KeyError):
        io4dolfinx.read_function(filename, u, time=0, name="nonexisting_name", backend=backend)
    # assert e.value.args[0] == (
    #     f"nonexisting_name not found in {filename}. Did you mean one of {variables}?"
    # )


def test_read_timestamps(get_dtype, mesh_2D, tmp_path, backend):
    if backend == "adios":
        suffix = ".bp"
    else:
        suffix = ".h5"

    mesh = mesh_2D
    dtype = get_dtype(mesh.geometry.x.dtype, False)

    el = basix.ufl.element(
        "Lagrange",
        mesh.basix_cell(),
        1,
        shape=(mesh.geometry.dim,),
        dtype=mesh.geometry.x.dtype,
    )
    V = dolfinx.fem.functionspace(mesh, el)

    u = dolfinx.fem.Function(V, dtype=dtype, name="u")
    v = dolfinx.fem.Function(V, dtype=dtype, name="v")

    f_path = mesh.comm.bcast(tmp_path, root=0)
    filename = (f_path / "read_time_stamps").with_suffix(suffix)

    t_u = [0.1, 1.4]
    t_v = [0.45, 1.2]

    io4dolfinx.write_mesh(filename, mesh, backend=backend)
    io4dolfinx.write_function(filename, u, time=t_u[0], backend=backend)
    io4dolfinx.write_function(filename, v, time=t_v[0], backend=backend)
    io4dolfinx.write_function(filename, u, time=t_u[1], backend=backend)
    io4dolfinx.write_function(filename, v, time=t_v[1], backend=backend)

    timestamps_u = io4dolfinx.read_timestamps(
        comm=mesh.comm, filename=filename, function_name="u", backend=backend
    )
    timestamps_v = io4dolfinx.read_timestamps(
        comm=mesh.comm, filename=filename, function_name="v", backend=backend
    )
    assert np.allclose(timestamps_u, t_u)
    assert np.allclose(timestamps_v, t_v)



import typing
from collections import ChainMap

from mpi4py import MPI

import dolfinx
import ipyparallel as ipp
import numpy as np
import numpy.typing
import numpy.typing as npt
import pytest

import io4dolfinx


def find_backends():
    backends = []
    try:
        import adios2

        if adios2.is_built_with_mpi:
            backends.append("adios2")
    except ModuleNotFoundError:
        pass
    try:
        import h5py

        if h5py.get_config().mpi:
            backends.append("h5py")
    except ModuleNotFoundError:
        pass
    return backends


@pytest.fixture(params=find_backends(), scope="function")
def backend(request):
    yield request.param


@pytest.fixture(scope="module")
def cluster():
    cluster = ipp.Cluster(engines="mpi", n=2)
    rc = cluster.start_and_connect_sync()
    rc.wait_for_engines(n=2)
    yield rc
    cluster.stop_cluster_sync()


@pytest.fixture(scope="function")
def write_function(tmp_path):
    def _write_function(
        mesh,
        el,
        f,
        dtype,
        backend: typing.Literal["adios2", "h5py"],
        name="uh",
        append: bool = False,
    ) -> str:
        V = dolfinx.fem.functionspace(mesh, el)
        uh = dolfinx.fem.Function(V, dtype=dtype)
        uh.interpolate(f)
        uh.name = name
        el_hash = (
            io4dolfinx.utils.element_signature(V)
            .replace(" ", "")
            .replace(",", "")
            .replace("(", "")
            .replace(")", "")
            .replace("[", "")
            .replace("]", "")
        )
        # Consistent tmp dir across processes
        f_path = MPI.COMM_WORLD.bcast(tmp_path, root=0)
        file_hash = f"{el_hash}_{np.dtype(dtype).name}"
        if backend == "adios2":
            suffix = ".bp"
        else:
            suffix = ".h5"

        filename = (f_path / f"mesh_{file_hash}").with_suffix(suffix)
        if mesh.comm.size != 1:
            if not append:
                io4dolfinx.write_mesh(filename, mesh, backend=backend)
            io4dolfinx.write_function(filename, uh, time=0.0, backend=backend)
        else:
            if MPI.COMM_WORLD.rank == 0:
                if not append:
                    io4dolfinx.write_mesh(filename, mesh, backend=backend)
                io4dolfinx.write_function(filename, uh, time=0.0, backend=backend)

        return filename

    return _write_function


@pytest.fixture(scope="function")
def read_function():
    def _read_function(
        comm, el, f, path, dtype, backend: typing.Literal["adios2", "h5py"], name="uh"
    ):
        mesh = io4dolfinx.read_mesh(
            path,
            comm,
            ghost_mode=dolfinx.mesh.GhostMode.shared_facet,
            backend=backend,
        )
        V = dolfinx.fem.functionspace(mesh, el)
        v = dolfinx.fem.Function(V, dtype=dtype)
        v.name = name
        io4dolfinx.read_function(path, v, backend=backend)
        v_ex = dolfinx.fem.Function(V, dtype=dtype)
        v_ex.interpolate(f)

        res = np.finfo(dtype).resolution
        np.testing.assert_allclose(v.x.array, v_ex.x.array, atol=10 * res, rtol=10 * res)

    return _read_function


@pytest.fixture(scope="function")
def get_dtype():
    def _get_dtype(in_dtype: np.dtype, is_complex: bool):
        dtype: numpy.typing.DTypeLike
        if in_dtype == np.float32:
            if is_complex:
                dtype = np.complex64
            else:
                dtype = np.float32
        elif in_dtype == np.float64:
            if is_complex:
                dtype = np.complex128
            else:
                dtype = np.float64
        else:
            raise ValueError("Unsuported dtype")
        return dtype

    return _get_dtype


@pytest.fixture(scope="function")
def write_function_time_dep(tmp_path):
    def _write_function_time_dep(
        mesh, el, f0, f1, t0, t1, dtype, backend: typing.Literal["adios2", "h5py"]
    ) -> str:
        V = dolfinx.fem.functionspace(mesh, el)
        uh = dolfinx.fem.Function(V, dtype=dtype)
        uh.interpolate(f0)
        el_hash = (
            io4dolfinx.utils.element_signature(V)
            .replace(" ", "")
            .replace(",", "")
            .replace("(", "")
            .replace(")", "")
            .replace("[", "")
            .replace("]", "")
        )
        file_hash = f"{el_hash}_{np.dtype(dtype).name}"
        # Consistent tmp dir across processes
        f_path = MPI.COMM_WORLD.bcast(tmp_path, root=0)
        if backend == "adios2":
            suffix = ".bp"
        else:
            suffix = ".h5"
        filename = (f_path / f"mesh_{file_hash}").with_suffix(suffix)
        if mesh.comm.size != 1:
            io4dolfinx.write_mesh(filename, mesh, backend=backend)
            io4dolfinx.write_function(filename, uh, time=t0, backend=backend)
            uh.interpolate(f1)
            io4dolfinx.write_function(filename, uh, time=t1, backend=backend)

        else:
            if MPI.COMM_WORLD.rank == 0:
                io4dolfinx.write_mesh(filename, mesh, backend=backend)
                io4dolfinx.write_function(filename, uh, time=t0, backend=backend)
                uh.interpolate(f1)
                io4dolfinx.write_function(filename, uh, time=t1, backend=backend)

        return filename

    return _write_function_time_dep


@pytest.fixture(scope="function")
def read_function_time_dep():
    def _read_function_time_dep(
        comm, el, f0, f1, t0, t1, path, dtype, backend: typing.Literal["adios2", "h5py"]
    ):
        mesh = io4dolfinx.read_mesh(
            path,
            comm,
            ghost_mode=dolfinx.mesh.GhostMode.shared_facet,
            backend=backend,
        )
        V = dolfinx.fem.functionspace(mesh, el)
        v = dolfinx.fem.Function(V, dtype=dtype)

        io4dolfinx.read_function(path, v, time=t1, backend=backend)
        v_ex = dolfinx.fem.Function(V, dtype=dtype)
        v_ex.interpolate(f1)

        res = np.finfo(dtype).resolution
        assert np.allclose(v.x.array, v_ex.x.array, atol=10 * res, rtol=10 * res)

        io4dolfinx.read_function(path, v, time=t0, backend=backend)
        v_ex = dolfinx.fem.Function(V, dtype=dtype)
        v_ex.interpolate(f0)

        res = np.finfo(dtype).resolution
        assert np.allclose(v.x.array, v_ex.x.array, atol=10 * res, rtol=10 * res)

    return _read_function_time_dep


def _generate_reference_map(
    mesh: dolfinx.mesh.Mesh,
    meshtag: dolfinx.mesh.MeshTags,
    comm: MPI.Intracomm,
    root: int,
) -> typing.Optional[dict[str, tuple[int, npt.NDArray]]]:
    """
    Helper function to generate map from meshtag value to its corresponding index and midpoint.

    Args:
        mesh: The mesh
        meshtag: The associated meshtag
        comm: MPI communicator to gather the map from all processes with
        root (int): Rank to store data on
    Returns:
        Root rank returns the map, all other ranks return None
    """
    mesh.topology.create_connectivity(meshtag.dim, mesh.topology.dim)
    midpoints = dolfinx.mesh.compute_midpoints(mesh, meshtag.dim, meshtag.indices)
    e_map = mesh.topology.index_map(meshtag.dim)
    value_to_midpoint = {}
    for index, value in zip(meshtag.indices, meshtag.values):
        value_to_midpoint[value] = (
            int(e_map.local_range[0] + index),
            midpoints[index],
        )
    global_map = comm.gather(value_to_midpoint, root=root)
    if comm.rank == root:
        return dict(ChainMap(*global_map))  # type: ignore
    return None


@pytest.fixture
def generate_reference_map():
    return _generate_reference_map



from mpi4py import MPI

import dolfinx
import numpy as np
import ufl

import io4dolfinx


def test_xdmf_mesh(tmp_path):
    tmp_path = MPI.COMM_WORLD.bcast(tmp_path, root=0)
    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 10, 10)
    tmp_file = tmp_path / "mesh.xdmf"
    with dolfinx.io.XDMFFile(MPI.COMM_WORLD, tmp_file, "w") as xdmf:
        xdmf.write_mesh(mesh)

    MPI.COMM_WORLD.barrier()
    in_grid = io4dolfinx.read_mesh(tmp_file, MPI.COMM_WORLD, backend="xdmf")

    assert mesh.topology.dim == in_grid.topology.dim
    assert mesh.geometry.dim == in_grid.geometry.dim
    for i in range(mesh.topology.dim):
        mesh.topology.create_entities(i)
        o_map = mesh.topology.index_map(i)
        in_grid.topology.create_entities(i)
        map = in_grid.topology.index_map(i)
        assert o_map.size_global == map.size_global

    assert mesh.geometry.index_map().size_global == in_grid.geometry.index_map().size_global
    org_area = mesh.comm.allreduce(
        dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * ufl.dx(domain=mesh))), op=MPI.SUM
    )
    area = mesh.comm.allreduce(
        dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * ufl.dx(domain=in_grid))), op=MPI.SUM
    )
    assert np.isclose(area, org_area)


def test_xdmf_function(tmp_path):
    tmp_path = MPI.COMM_WORLD.bcast(tmp_path, root=0)
    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 8, 10)

    def f(x):
        return (x[0], x[1], -2 * x[1], 3 * x[0])

    V = dolfinx.fem.functionspace(mesh, ("Lagrange", 1, (4,)))
    u = dolfinx.fem.Function(V, name="u")
    u.interpolate(f)

    tmp_file = tmp_path / "function.xdmf"
    with dolfinx.io.XDMFFile(MPI.COMM_WORLD, tmp_file, "w") as xdmf:
        xdmf.write_mesh(mesh)
        xdmf.write_function(u)

    MPI.COMM_WORLD.barrier()

    in_grid = io4dolfinx.read_mesh(tmp_file, MPI.COMM_WORLD, backend="xdmf")
    names = io4dolfinx.read_function_names(
        tmp_file, MPI.COMM_WORLD, backend_args={}, backend="xdmf"
    )
    assert len(names) == 1
    assert names[0] == "u"

    u = io4dolfinx.read_point_data(tmp_file, "u", in_grid, backend="xdmf")
    u_ref = dolfinx.fem.Function(u.function_space)
    u_ref.interpolate(f)
    eps = np.finfo(mesh.geometry.x.dtype).eps
    np.testing.assert_allclose(u.x.array, u_ref.x.array, atol=eps)



"""
Module that uses DOLFINx/H5py to import XDMF files.
"""

from pathlib import Path
from typing import Any
from xml.etree import ElementTree

from mpi4py import MPI

import basix
import dolfinx
import numpy as np
import numpy.typing as npt

from io4dolfinx.structures import ArrayData, FunctionData, MeshData, MeshTagsData, ReadMeshData
from io4dolfinx.utils import check_file_exists, compute_local_range

from .. import FileMode, ReadMode
from ..h5py.backend import h5pyfile

read_mode = ReadMode.parallel


def extract_function_names_and_timesteps(filename: Path | str) -> dict[str, list[str]]:
    tree = ElementTree.parse(filename)
    root = tree.getroot()
    mesh_nodes = root.findall(".//Grid[@CollectionType='Temporal']")
    function_names = []
    for mesh in mesh_nodes:
        function_names.append(mesh.attrib["Name"])

    time_stamps: dict[str, list[str]] = {name: [] for name in function_names}
    for name in function_names:
        time_steps = root.findall(f".//Grid[@Name='{name}']")
        for time in time_steps:
            step = time.find(".//Time")
            if step is not None:
                val = step.attrib["Value"]
                time_stamps[name].append(val)
    for name in function_names:
        float_steps = np.argsort(np.array(list(set(time_stamps[name])), dtype=np.float64))
        time_stamps[name] = np.array(list(set(time_stamps[name])), dtype=str)[float_steps].tolist()
    return time_stamps


def get_default_backend_args(arguments: dict[str, Any] | None) -> dict[str, Any]:
    """Get default backend arguments given a set of input arguments.

    Args:
        arguments: Input backend arguments

    Returns:
        Updated backend arguments
    """
    args = arguments or {}
    return args


def read_mesh_data(
    filename: Path | str,
    comm: MPI.Intracomm,
    time: str | float | None,
    read_from_partition: bool,
    backend_args: dict[str, Any] | None,
) -> ReadMeshData:
    """Read mesh data from file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        time: Time stamp associated with the mesh to read
        read_from_partition: Whether to read partition information
        backend_args: Arguments to backend

    Returns:
        Internal data structure for the mesh data read from file
    """
    assert not read_from_partition
    check_file_exists(filename)
    with dolfinx.io.XDMFFile(comm, filename, "r") as file:
        cell_shape, cell_degree = file.read_cell_type()
        cells = file.read_topology_data()
        x = file.read_geometry_data()
    return ReadMeshData(
        cells=cells,
        cell_type=cell_shape.name,
        x=x,
        lvar=int(basix.LagrangeVariant.equispaced),
        degree=cell_degree,
    )


def read_point_data(
    filename: Path | str,
    name: str,
    comm: MPI.Intracomm,
    time: float | str | None,
    backend_args: dict[str, Any] | None,
) -> tuple[np.ndarray, int]:
    """Read data from the nodes of a mesh.

    Args:
        filename: Path to file
        name: Name of point data
        comm: Communicator to launch IO on.
        time: The time stamp
        backend_args: The backend arguments

    Returns:
       Data local to process (contiguous, no mpi comm) and local start range
    """
    # Find function with name u in xml tree
    check_file_exists(filename)
    filename = Path(filename)

    tree = ElementTree.parse(filename)
    root = tree.getroot()
    backend_args = get_default_backend_args(backend_args)
    if time is not None:
        time_steps = root.findall(f".//Grid[@Name='{name}']")
        time_found = False
        for time_node in time_steps:
            step_node = time_node.find(".//Time")
            assert isinstance(step_node, ElementTree.Element)
            if step_node.attrib["Value"] == time:
                time_found = True
                break
        func_node = time_node.find(f".//Attribute[@Name='{name}']")
        if not time_found:
            raise RuntimeError(f"Function {name} at time={time} not found in {filename}")
    else:
        func_node = root.find(f".//Attribute[@Name='{name}']")
    assert isinstance(func_node, ElementTree.Element)
    data_node = func_node.find(".//DataItem")
    assert isinstance(data_node, ElementTree.Element)
    global_shape = data_node.attrib["Dimensions"].split(" ")
    func_path = data_node.text
    assert isinstance(func_path, str)
    data_file, data_loc = func_path.split(":")
    data_path = filename.parent / data_file
    with h5pyfile(data_path, "r", comm=comm) as h5file:
        data = h5file[data_loc]
        for s1, s2 in zip(data.shape, global_shape, strict=True):
            assert int(s1) == int(s2)
        lr = compute_local_range(comm, data.shape[0])
        local_range_start = lr[0]
        dataset = data[slice(*lr), :]
    return dataset, local_range_start


def read_attributes(
    filename: Path | str,
    comm: MPI.Intracomm,
    name: str,
    backend_args: dict[str, Any] | None,
) -> dict[str, Any]:
    """Read attributes from file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the attribute group
        backend_args: Arguments to backend

    Returns:
        Dictionary of attributes read from file
    """
    raise NotImplementedError("The XDMF backend cannot read attributes.")


def read_timestamps(
    filename: Path | str,
    comm: MPI.Intracomm,
    function_name: str,
    backend_args: dict[str, Any] | None,
) -> npt.NDArray[np.float64 | str]:  # type: ignore[type-var]
    """Read timestamps from file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        function_name: Name of the function to read timestamps for
        backend_args: Arguments to backend

    Returns:
        Numpy array of timestamps read from file
    """
    tree = ElementTree.parse(filename)
    root = tree.getroot()
    time_stamps = []
    time_steps = root.findall(f".//Grid[@Name='{function_name}']")
    for time in time_steps:
        step = time.find(".//Time")
        if step is not None:
            val = step.attrib["Value"]
            time_stamps.append(val)
    float_steps = np.argsort(np.array(list(set(time_stamps)), dtype=np.float64))
    return np.array(list(set(time_stamps)), dtype=str)[float_steps].tolist()


def write_attributes(
    filename: Path | str,
    comm: MPI.Intracomm,
    name: str,
    attributes: dict[str, np.ndarray],
    backend_args: dict[str, Any] | None,
):
    """Write attributes to file.

    Args:
        filename: Path to file to write to
        comm: MPI communicator used in storage
        name: Name of the attribute group
        attributes: Dictionary of attributes to write
        backend_args: Arguments to backend
    """
    raise NotImplementedError("The XDMF backend cannot write attributes.")


def write_mesh(
    filename: Path | str,
    comm: MPI.Intracomm,
    mesh: MeshData,
    backend_args: dict[str, Any] | None,
    mode: FileMode,
    time: float,
):
    """
    Write a mesh to file.

    Args:
        comm: MPI communicator used in storage
        mesh: Internal data structure for the mesh data to save to file
        filename: Path to file to write to
        backend_args: Arguments to backend
        mode: File-mode to store the mesh
        time: Time stamp associated with the mesh
    """
    raise NotImplementedError("The XDMF backend cannot write meshes.")


def write_meshtags(
    filename: str | Path,
    comm: MPI.Intracomm,
    data: MeshTagsData,
    backend_args: dict[str, Any] | None,
):
    """Write mesh tags to file.

    Args:
        filename: Path to file to write to
        comm: MPI communicator used in storage
        data: Internal data structure for the mesh tags to save to file
        backend_args: Arguments to backend
    """
    raise NotImplementedError("The XDMF backend cannot write meshtags.")


def read_meshtags_data(
    filename: str | Path,
    comm: MPI.Intracomm,
    name: str,
    backend_args: dict[str, Any] | None,
) -> MeshTagsData:
    """Read mesh tags from file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the mesh tags to read
        backend_args: Arguments to backend

    Returns:
        Internal data structure for the mesh tags read from file
    """
    raise NotImplementedError("The XDMF backend cannot read meshtags.")


def read_dofmap(
    filename: str | Path,
    comm: MPI.Intracomm,
    name: str,
    backend_args: dict[str, Any] | None,
) -> dolfinx.graph.AdjacencyList:
    """Read the dofmap of a function with a given name.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the function to read the dofmap for
        backend_args: Arguments to backend

    Returns:
        Dofmap as an AdjacencyList
    """
    raise NotImplementedError("The XDMF backend cannot make checkpoints.")


def read_dofs(
    filename: str | Path,
    comm: MPI.Intracomm,
    name: str,
    time: float,
    backend_args: dict[str, Any] | None,
) -> tuple[npt.NDArray[np.float32 | np.float64 | np.complex64 | np.complex128], int]:
    """Read the dofs (values) of a function with a given name from a given timestep.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the function to read the dofs for
        time: Time stamp associated with the function to read
        backend_args: Arguments to backend

    Returns:
        Contiguous sequence of degrees of freedom (with respect to input data)
        and the global starting point on the process.
        Process 0 has [0, M), process 1 [M, N), process 2 [N, O) etc.
    """
    raise NotImplementedError("The XDMF backend cannot make checkpoints.")


def read_cell_perms(
    comm: MPI.Intracomm, filename: Path | str, backend_args: dict[str, Any] | None
) -> npt.NDArray[np.uint32]:
    """
    Read cell permutation from file with given communicator,
    Split in continuous chunks based on number of cells in the input data.

    Args:
        comm: MPI communicator used in storage
        filename: Path to file to read from
        backend_args: Arguments to backend

    Returns:
        Contiguous sequence of permutations (with respect to input data)
        Process 0 has [0, M), process 1 [M, N), process 2 [N, O) etc.
    """
    raise NotImplementedError("The XDMF backend cannot make checkpoints.")


def write_function(
    filename: Path,
    comm: MPI.Intracomm,
    u: FunctionData,
    time: float,
    mode: FileMode,
    backend_args: dict[str, Any] | None,
):
    """
    Write a function to file.

    Args:
        comm: MPI communicator used in storage
        u: Internal data structure for the function data to save to file
        filename: Path to file to write to
        time: Time stamp associated with function
        mode: File-mode to store the function
        backend_args: Arguments to backend
    """
    raise NotImplementedError("The XDMF backend cannot make checkpoints.")


def read_legacy_mesh(
    filename: Path | str, comm: MPI.Intracomm, group: str
) -> tuple[npt.NDArray[np.int64], npt.NDArray[np.floating], str | None]:
    """Read in the mesh topology, geometry and (optionally) cell type from a
    legacy DOLFIN HDF5-file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        group: Group in HDF5 file where mesh is stored

    Returns:
        Tuple containing:
            - Topology as a (num_cells, num_vertices_per_cell) array of global vertex indices
            - Geometry as a (num_vertices, geometric_dimension) array of vertex coordinates
            - Cell type as a string (e.g. "tetrahedron") or None if not found
    """
    raise NotImplementedError("The XDMF backend cannot read legacy DOLFIN meshes.")


def snapshot_checkpoint(
    filename: Path | str,
    mode: FileMode,
    u: dolfinx.fem.Function,
    backend_args: dict[str, Any] | None,
):
    """Create a snapshot checkpoint of a dolfinx function.

    Args:
        filename: Path to file to read from
        mode: File-mode to store the function
        u: dolfinx function to create a snapshot checkpoint for
        backend_args: Arguments to backend
    """
    raise NotImplementedError("The XDMF backend cannot make checkpoints.")


def read_hdf5_array(
    comm: MPI.Intracomm,
    filename: Path | str,
    group: str,
    backend_args: dict[str, Any] | None,
) -> tuple[np.ndarray, int]:
    """Read an array from an HDF5 file.

    Args:
        comm: MPI communicator used in storage
        filename: Path to file to read from
        group: Group in HDF5 file where array is stored
        backend_args: Arguments to backend

    Returns:
        Tuple containing:
            - Numpy array read from file
            - Global starting point on the process.
                Process 0 has [0, M), process 1 [M, N), process 2 [N, O) etc.
    """
    raise NotImplementedError("The XDMF backend cannot read HDF5 arrays")


def read_function_names(
    filename: Path | str, comm: MPI.Intracomm, backend_args: dict[str, Any] | None
) -> list[str]:
    """Read all function names from a file.

    Args:
        filename: Path to file
        comm: MPI communicator to launch IO on.
        backend_args: Arguments to backend

    Returns:
        A list of function names.
    """
    # Find all functions in xml tree
    check_file_exists(filename)
    filename = Path(filename)

    tree = ElementTree.parse(filename)
    root = tree.getroot()
    backend_args = get_default_backend_args(backend_args)
    # Functions in checkpoint format
    checkpoint_funcs = root.findall(".//Attribute[@ItemType='FiniteElementFunction']")
    names = [func.attrib["Name"] for func in checkpoint_funcs]
    # Temporal funcs
    temporal_funcs = root.findall(".//Grid[@GridType='Collection']")
    for func in temporal_funcs:
        names.append(func.attrib["Name"])
    return list(set(names))


def read_cell_data(
    filename: Path | str,
    name: str,
    comm: MPI.Intracomm,
    time: str | float | None,
    backend_args: dict[str, Any] | None,
) -> tuple[npt.NDArray[np.int64], np.ndarray]:
    """Read data from the cells of a mesh.

    Args:
        filename: Path to file
        name: Name of point data
        comm: Communicator to launch IO on.
        time: The time stamp
        backend_args: The backend arguments
    Returns:
        A tuple (topology, dofs) where topology contains the
        vertex indices of the cells, dofs the degrees of
        freedom within that cell.
    """
    # Find function with name u in xml tree
    check_file_exists(filename)
    filename = Path(filename)

    tree = ElementTree.parse(filename)
    root = tree.getroot()
    backend_args = get_default_backend_args(backend_args)
    if time is not None:
        time_steps = root.findall(f".//Grid[@Name='{name}']")
        time_found = False
        for time_node in time_steps:
            step_node = time_node.find(".//Time")
            assert isinstance(step_node, ElementTree.Element)
            if np.isclose(float(step_node.attrib["Value"]), time):
                time_found = True
                break
        func_node = time_node.find(f".//Attribute[@Name='{name}']")
        if not time_found:
            raise RuntimeError(f"Function {name} at time={time} not found in {filename}")
    else:
        func_node = root.find(f".//Attribute[@Name='{name}']")
    assert func_node is not None
    if func_node.attrib["ItemType"] == "FiniteElementFunction":
        if (family := func_node.attrib["ElementFamily"]) != "DG" or (
            degree := int(func_node.attrib["ElementDegree"])
        ) != 0:
            raise ValueError(
                f"Cannot read in finite element function ({family}, {degree}) as cell data."
            )
        # Get vector sub-element
        vec_el = None
        for node in func_node.iter():
            comp = node.text
            assert comp is not None
            if "vector" in comp:
                vec_el = node
                break
        assert vec_el is not None
        dof_dimensions = np.array(vec_el.attrib["Dimensions"].split(" "), dtype=np.int32)
        vtxt = vec_el.text
        assert vtxt is not None
        vector_file, vector_h5path = vtxt.split(":")

        grid_node = root.find(f".//Attribute[@Name='{name}']/..")
        assert grid_node is not None
        topology = grid_node.find("./Topology")
        assert topology is not None
        ttext = topology[0].text
        assert ttext is not None
        topology_file, topology_h5path = ttext.split(":")

        with h5pyfile(filename.parent / vector_file, "r", comm=comm) as h5_mesh:
            data_loc = h5_mesh[vector_h5path]
            data_shape = data_loc.shape[0]
            assert int(np.prod(data_shape)) == int(np.prod(dof_dimensions))
            local_range = compute_local_range(comm, data_shape)
            vec_dofs = data_loc[slice(*local_range)]

        with h5pyfile(filename.parent / topology_file, "r", comm=comm) as h5_top:
            data_loc = h5_top[topology_h5path]
            top_data_shape = data_loc.shape[0]
            assert dof_dimensions[0] == top_data_shape
            local_range = compute_local_range(comm, data_shape)
            top_dofs = data_loc[slice(*local_range)].astype(np.int64)

        return top_dofs, vec_dofs
    else:
        raise NotImplementedError("Not implemented yet.")


def write_data(
    filename: Path | str,
    point_data: ArrayData,
    comm: MPI.Intracomm,
    time: str | float | None,
    mode: FileMode,
    backend_args: dict[str, Any] | None,
):
    """Write a 2D-array to file (distributed across proceses with MPI).


    Args:
        filename: Path to file
        point_data: Data to write to file
        time: Time stamp
        mode: Append or write
        backend_args: The backend arguments
    """
    raise NotImplementedError("XDMF has not implemented this yet")



# Copyright (C) 2024 Jørgen Schartum Dokken
#
# This file is part of io4dolfinx
#
# SPDX-License-Identifier:    MIT

from __future__ import annotations

from dataclasses import dataclass

import numpy as np
import numpy.typing as npt
from dolfinx.graph import AdjacencyList

"""Internal library classes for storing mesh and function data"""
__all__ = ["MeshData", "FunctionData", "ReadMeshData", "MeshTagsData"]


@dataclass
class MeshData:
    """
    Container for distributed mesh data that will be stored to disk
    """

    #: Two-dimensional array of node coordinates
    local_geometry: npt.NDArray[np.float32] | npt.NDArray[np.float64]
    local_geometry_pos: tuple[int, int]  #: Insert range on current process for geometry nodes
    num_nodes_global: int  #: Number of nodes in global geometry array

    local_topology: npt.NDArray[np.int64]  #: Two-dimensional connectivity array for mesh topology
    #: Insert range on current process for topology
    local_topology_pos: tuple[int, int]
    num_cells_global: int  #: NUmber of cells in global topology

    cell_type: str  #: The cell type
    degree: int  #: Degree of underlying Lagrange element
    lagrange_variant: int  #: Lagrange-variant of DOFs

    # Partitioning_information
    store_partition: bool  #: Indicator if one should store mesh partitioning
    partition_processes: int | None = None  #: Number of processes in partition
    ownership_array: npt.NDArray[np.int32] | None = None  #: Ownership array for cells
    ownership_offset: npt.NDArray[np.int32] | None = None  #: Ownership offset for cells
    partition_range: tuple[int, int] | None = (
        None  #: Local insert position for partitioning information
    )
    partition_global: int | None = None  #: Global size of partitioning array


@dataclass
class FunctionData:
    """
    Container for distributed function data that will be written to file
    """

    cell_permutations: npt.NDArray[np.uint32]  #: Cell permutations for dofmap
    local_cell_range: tuple[int, int]  #: Range of cells on current process
    num_cells_global: int  #: Number of cells in global topology
    dofmap_array: npt.NDArray[np.int64]  #: Local function dofmap (using global indices)
    dofmap_offsets: npt.NDArray[np.int64]  #: Global dofmap offsets
    dofmap_range: tuple[int, int]  #: Range of dofmap on current process
    global_dofs_in_dofmap: int  #: Number of entries in global dofmap
    values: npt.NDArray[np.floating]  #: Local function values
    dof_range: tuple[int, int]  #: Range of local function values
    num_dofs_global: int  #: Number of global function values
    name: str  #: Name of function


@dataclass
class ReadMeshData:
    """Container containing data that will be read into DOLFINx"""

    #: Two-dimensional array containing global cell->node connectivity
    cells: npt.NDArray[np.int64]
    cell_type: str  #: The cell type of the mesh
    x: npt.NDArray[np.floating]  #: The mesh coordinates
    lvar: int  #: The Lagrange variant
    degree: int  #: The degree of the underlying Lagrange element
    #: Partitioning information per cell on the process
    partition_graph: AdjacencyList | None = None


@dataclass
class MeshTagsData:
    name: str  #: Name of tag
    values: npt.NDArray  # Array of values
    indices: npt.NDArray[np.int64]  # Global indices of the entities
    dim: int  # Topological dimension of the entities

    # Optional entries (used for writing to disk)
    num_entities_global: int | None = None  #: Global number of entities that will be written out
    num_dofs_per_entity: int | None = None  #: Number of dofs per entity
    #: Starting index in output array `(0<=local_start<num_entities_global)``
    local_start: int | None = None

    # Optional info to help visualization
    cell_type: str | None = None  #: The cell type


@dataclass
class ArrayData:
    name: str  # Name of point data
    values: npt.NDArray  # Array of values (num_entities, num_components)
    global_shape: tuple[int, int]  # Global shape of data
    local_range: tuple[int, int]  # Start and stop of local array in global array
    type: str  # Type of array-data



import importlib

from mpi4py import MPI

import dolfinx
import numpy as np
import pytest
import ufl

from io4dolfinx import FileMode, read_mesh, write_mesh


def get_hdf5_version():
    """Get the HDF5 library version found on the system.

    Note: This function first tries to get the version via h5py.
    If h5py is not installed, it attempts to load the HDF5 shared library
    directly using ctypes to query the version.
    """
    import ctypes
    from ctypes import byref, c_uint

    from packaging.version import parse as _v

    try:
        import h5py

        return _v(h5py.version.hdf5_version)
    except ImportError:
        try:
            # Try to load default HDF5 library names
            # If adios2 is already imported, the symbols might be globally available
            try:
                libhdf5 = ctypes.CDLL("libhdf5.so")
            except OSError:
                try:
                    libhdf5 = ctypes.CDLL("libhdf5.dylib")  # MacOS
                except OSError:
                    print("\n[Method 2] Could not load libhdf5 shared library directly.")
                    libhdf5 = None

            if libhdf5:
                major = c_uint()
                minor = c_uint()
                rel = c_uint()

                # Call the C function
                libhdf5.H5get_libversion(byref(major), byref(minor), byref(rel))
                return _v(f"{major.value}.{minor.value}.{rel.value}")
        except Exception:
            pass

    except Exception:
        pass

    raise RuntimeError("Failed to get HDF5 version")


@pytest.mark.parametrize(
    "backend, encoder, suffix",
    [
        ("adios2", "BP4", ".bp"),
        ("adios2", "HDF5", ".h5"),
        ("adios2", "BP5", ".bp"),
        ("h5py", "HDF5", ".h5"),
    ],
)
@pytest.mark.parametrize(
    "ghost_mode", [dolfinx.mesh.GhostMode.shared_facet, dolfinx.mesh.GhostMode.none]
)
@pytest.mark.parametrize("store_partition", [True, False])
def test_mesh_read_writer(backend, encoder, suffix, ghost_mode, tmp_path, store_partition):
    N = 7

    if backend == "adios2" and encoder == "HDF5" and get_hdf5_version().major >= 2:
        pytest.skip("HDF5 version >= 2 is not supported due to ADIOS2 limitations.")
    try:
        importlib.import_module(backend)
    except ModuleNotFoundError:
        pytest.skip(f"{backend} not installed")

    # Consistent tmp dir across processes
    fname = MPI.COMM_WORLD.bcast(tmp_path, root=0)
    file = fname / f"{backend}_mesh_{encoder}_{store_partition}"
    xdmf_file = fname / f"xdmf_mesh_{encoder}_{ghost_mode}_{store_partition}"

    mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, N, N, N, ghost_mode=ghost_mode)

    backend_args = None
    if backend == "adios2":
        backend_args = {"engine": encoder}
    write_mesh(
        file.with_suffix(suffix),
        mesh,
        store_partition_info=store_partition,
        backend_args=backend_args,
        backend=backend,
    )

    mesh.comm.Barrier()
    with dolfinx.io.XDMFFile(mesh.comm, xdmf_file.with_suffix(".xdmf"), "w") as xdmf:
        xdmf.write_mesh(mesh)
    mesh.comm.Barrier()

    in_mesh = read_mesh(
        file.with_suffix(suffix),
        MPI.COMM_WORLD,
        ghost_mode=ghost_mode,
        read_from_partition=store_partition,
        backend_args=backend_args,
        backend=backend,
    )
    in_mesh.comm.Barrier()
    if store_partition:

        def compute_distance_matrix(points_A, points_B, tol=1e-12):
            points_A_e = np.expand_dims(points_A, 1)
            points_B_e = np.expand_dims(points_B, 0)
            distances = np.sum(np.square(points_A_e - points_B_e), axis=2)
            return distances < tol

        cell_map = mesh.topology.index_map(mesh.topology.dim)
        new_cell_map = in_mesh.topology.index_map(in_mesh.topology.dim)
        assert cell_map.size_local == new_cell_map.size_local
        assert cell_map.num_ghosts == new_cell_map.num_ghosts
        mesh.topology.create_connectivity(mesh.topology.dim, mesh.topology.dim)
        midpoints = dolfinx.mesh.compute_midpoints(
            mesh,
            mesh.topology.dim,
            np.arange(cell_map.size_local + cell_map.num_ghosts, dtype=np.int32),
        )
        in_mesh.topology.create_connectivity(in_mesh.topology.dim, in_mesh.topology.dim)
        new_midpoints = dolfinx.mesh.compute_midpoints(
            in_mesh,
            in_mesh.topology.dim,
            np.arange(new_cell_map.size_local + new_cell_map.num_ghosts, dtype=np.int32),
        )
        # Check that all points in owned by initial mesh is owned by the new mesh
        # (might be locally reordered)
        owned_distances = compute_distance_matrix(
            midpoints[: cell_map.size_local], new_midpoints[: new_cell_map.size_local]
        )
        np.testing.assert_allclose(np.sum(owned_distances, axis=1), 1)
        # Check that all points that are ghosted in original mesh is ghosted on the
        # same process in the new mesh
        ghost_distances = compute_distance_matrix(
            midpoints[cell_map.size_local :], new_midpoints[new_cell_map.size_local :]
        )
        np.testing.assert_allclose(np.sum(ghost_distances, axis=1), 1)

    mesh.comm.Barrier()

    with dolfinx.io.XDMFFile(mesh.comm, xdmf_file.with_suffix(".xdmf"), "r") as xdmf:
        mesh_xdmf = xdmf.read_mesh(ghost_mode=ghost_mode)

    for i in range(mesh.topology.dim + 1):
        mesh.topology.create_entities(i)
        mesh_xdmf.topology.create_entities(i)
        in_mesh.topology.create_entities(i)
        assert (
            mesh_xdmf.topology.index_map(i).size_global == in_mesh.topology.index_map(i).size_global
        )

    # Check that integration over different entities are consistent
    measures = (
        [ufl.ds, ufl.dx] if ghost_mode is dolfinx.mesh.GhostMode.none else [ufl.ds, ufl.dS, ufl.dx]
    )
    for measure in measures:
        form = dolfinx.fem.form(1 * measure(domain=in_mesh))
        c_adios = dolfinx.fem.assemble_scalar(form)
        c_ref = dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * measure(domain=mesh)))
        c_xdmf = dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * measure(domain=mesh_xdmf)))
        assert np.isclose(
            in_mesh.comm.allreduce(c_adios, MPI.SUM),
            mesh.comm.allreduce(c_xdmf, MPI.SUM),
        )
        assert np.isclose(
            in_mesh.comm.allreduce(c_adios, MPI.SUM),
            mesh.comm.allreduce(c_ref, MPI.SUM),
        )


@pytest.mark.parametrize(
    "backend, encoder, suffix",
    [("adios2", "BP4", ".bp"), ("adios2", "BP5", ".bp"), ("h5py", None, ".h5")],
)
@pytest.mark.parametrize(
    "ghost_mode", [dolfinx.mesh.GhostMode.shared_facet, dolfinx.mesh.GhostMode.none]
)
@pytest.mark.parametrize("store_partition", [True, False])
def test_timedep_mesh(encoder, backend, suffix, ghost_mode, tmp_path, store_partition):
    try:
        importlib.import_module(backend)
    except ModuleNotFoundError:
        pytest.skip(f"{backend} not installed")

    # Currently unsupported, unclear why ("HDF5", ".h5"),
    N = 13
    # Consistent tmp dir across processes
    fname = MPI.COMM_WORLD.bcast(tmp_path, root=0)
    file = fname / f"adios_time_dep_mesh_{encoder}"
    mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, N, N, N, ghost_mode=ghost_mode)

    def u(x):
        return np.asarray([x[0] + 0.1 * np.sin(x[1]), 0.2 * np.cos(x[1]), x[2]])

    backend_args = None
    if backend == "adios2":
        backend_args = {"engine": encoder}
    write_mesh(
        file.with_suffix(suffix),
        mesh,
        mode=FileMode.write,
        time=0.0,
        store_partition_info=store_partition,
        backend_args=backend_args,
        backend=backend,
    )
    delta_x = u(mesh.geometry.x.T).T
    mesh.geometry.x[:] += delta_x
    write_mesh(
        file.with_suffix(suffix),
        mesh,
        mode=FileMode.append,
        time=3.0,
        backend_args=backend_args,
        backend=backend,
    )
    mesh.geometry.x[:] -= delta_x

    mesh_first = read_mesh(
        file.with_suffix(suffix),
        MPI.COMM_WORLD,
        ghost_mode,
        time=0.0,
        read_from_partition=store_partition,
        backend_args=backend_args,
        backend=backend,
    )
    mesh_first.comm.Barrier()

    # Check that integration over different entities are consistent
    measures = [ufl.ds, ufl.dx]
    if ghost_mode == dolfinx.mesh.GhostMode.shared_facet:
        measures.append(ufl.dx)
    for measure in measures:
        # try:
        c_adios = dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * measure(domain=mesh_first)))
        c_ref = dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * measure(domain=mesh)))
        assert np.isclose(
            mesh_first.comm.allreduce(c_adios, MPI.SUM),
            mesh.comm.allreduce(c_ref, MPI.SUM),
        )

    mesh.geometry.x[:] += delta_x
    mesh_second = read_mesh(
        file.with_suffix(suffix),
        MPI.COMM_WORLD,
        ghost_mode,
        time=3.0,
        read_from_partition=store_partition,
        backend_args=backend_args,
        backend=backend,
    )
    mesh_second.comm.Barrier()
    measures = [ufl.ds, ufl.dx]
    if ghost_mode == dolfinx.mesh.GhostMode.shared_facet:
        measures.append(ufl.dx)
    for measure in measures:
        c_adios = dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * measure(domain=mesh_second)))
        c_ref = dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * measure(domain=mesh)))
        assert np.isclose(
            mesh_second.comm.allreduce(c_adios, MPI.SUM),
            mesh.comm.allreduce(c_ref, MPI.SUM),
        )



from __future__ import annotations

import itertools
import os
import typing
from collections.abc import Callable
from pathlib import Path

from mpi4py import MPI

import basix
import basix.ufl
import dolfinx
import numpy as np
import pytest

import io4dolfinx

dtypes = [np.float64, np.float32]  # Mesh geometry dtypes

two_dimensional_cell_types = [
    dolfinx.mesh.CellType.triangle,
    dolfinx.mesh.CellType.quadrilateral,
]
three_dimensional_cell_types = [
    dolfinx.mesh.CellType.tetrahedron,
    dolfinx.mesh.CellType.hexahedron,
]

two_dim_combinations = itertools.product(dtypes, two_dimensional_cell_types)
three_dim_combinations = itertools.product(dtypes, three_dimensional_cell_types)


@pytest.fixture(scope="module")
def create_simplex_mesh_2D(tmp_path_factory):
    mesh = dolfinx.mesh.create_unit_square(
        MPI.COMM_WORLD,
        10,
        10,
        cell_type=dolfinx.mesh.CellType.triangle,
        dtype=np.float64,
    )
    fname = tmp_path_factory.mktemp("output") / "original_mesh_2D_simplex.xdmf"
    fname = MPI.COMM_WORLD.bcast(fname, root=0)
    with dolfinx.io.XDMFFile(MPI.COMM_WORLD, fname, "w") as xdmf:
        xdmf.write_mesh(mesh)
    return fname


@pytest.fixture(scope="module")
def create_simplex_mesh_3D(tmp_path_factory):
    mesh = dolfinx.mesh.create_unit_cube(
        MPI.COMM_WORLD,
        5,
        5,
        5,
        cell_type=dolfinx.mesh.CellType.tetrahedron,
        dtype=np.float64,
    )
    fname = tmp_path_factory.mktemp("output") / "original_mesh_3D_simplex.xdmf"
    with dolfinx.io.XDMFFile(MPI.COMM_WORLD, fname, "w") as xdmf:
        xdmf.write_mesh(mesh)
    return fname


@pytest.fixture(scope="module")
def create_non_simplex_mesh_2D(tmp_path_factory):
    mesh = dolfinx.mesh.create_unit_square(
        MPI.COMM_WORLD,
        10,
        10,
        cell_type=dolfinx.mesh.CellType.quadrilateral,
        dtype=np.float64,
    )
    fname = tmp_path_factory.mktemp("output") / "original_mesh_2D_non_simplex.xdmf"
    with dolfinx.io.XDMFFile(MPI.COMM_WORLD, fname, "w") as xdmf:
        xdmf.write_mesh(mesh)
    return fname


@pytest.fixture(scope="module")
def create_non_simplex_mesh_3D(tmp_path_factory):
    mesh = dolfinx.mesh.create_unit_cube(
        MPI.COMM_WORLD,
        5,
        5,
        5,
        cell_type=dolfinx.mesh.CellType.hexahedron,
        dtype=np.float64,
    )
    fname = tmp_path_factory.mktemp("output") / "original_mesh_3D_non_simplex.xdmf"
    with dolfinx.io.XDMFFile(MPI.COMM_WORLD, fname, "w") as xdmf:
        xdmf.write_mesh(mesh)
    return fname


@pytest.fixture(params=two_dim_combinations, scope="module")
def create_2D_mesh(request, tmpdir_factory):
    dtype, cell_type = request.param
    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 5, 7, cell_type=cell_type, dtype=dtype)
    fname = Path(tmpdir_factory.mktemp("output")) / f"original_mesh_2D_{dtype}_{cell_type}.xdmf"
    with dolfinx.io.XDMFFile(MPI.COMM_WORLD, fname, "w") as xdmf:
        xdmf.write_mesh(mesh)
    return fname


@pytest.fixture(params=three_dim_combinations, scope="module")
def create_3D_mesh(request, tmpdir_factory):
    dtype, cell_type = request.param
    mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, 5, 7, 3, cell_type=cell_type, dtype=dtype)
    fname = Path(tmpdir_factory.mktemp("output")) / f"original_mesh_3D_{dtype}_{cell_type}.xdmf"
    with dolfinx.io.XDMFFile(MPI.COMM_WORLD, fname, "w") as xdmf:
        xdmf.write_mesh(mesh)
    return fname


def write_function_original(
    write_mesh: bool,
    mesh: dolfinx.mesh.Mesh,
    el: basix.ufl._ElementBase,
    f: Callable[[np.ndarray], np.ndarray],
    dtype: np.dtype,
    name: str,
    path: Path,
    backend: typing.Literal["adios2", "h5py"],
) -> Path:
    """Convenience function for writing function to file on the original input mesh"""
    V = dolfinx.fem.functionspace(mesh, el)
    uh = dolfinx.fem.Function(V, dtype=dtype)
    uh.interpolate(f)
    uh.name = name
    el_hash = (
        io4dolfinx.utils.element_signature(V)
        .replace(" ", "")
        .replace(",", "")
        .replace("(", "")
        .replace(")", "")
        .replace("[", "")
        .replace("]", "")
    )

    file_hash = f"{el_hash}_{np.dtype(dtype).name}"
    if backend == "adios2":
        suffix = ".bp"
    elif backend == "h5py":
        suffix = ".h5"
    else:
        raise NotImplementedError(f"Unknown backend {backend}")

    filename = (path / f"mesh_{file_hash}").with_suffix(suffix)
    if write_mesh:
        io4dolfinx.write_mesh_input_order(filename, mesh, backend=backend)
    io4dolfinx.write_function_on_input_mesh(filename, uh, time=0.0, backend=backend)
    return filename


def read_function_original(
    mesh_fname: Path,
    u_fname: Path,
    u_name: str,
    family: str,
    degree: int,
    f: Callable[[np.ndarray], np.ndarray],
    u_dtype: np.dtype,
    backend: typing.Literal["adios2", "h5py"],
):
    """
    Convenience function for reading mesh with IPython-parallel and compare to exact solution
    """
    from mpi4py import MPI

    import dolfinx

    import io4dolfinx

    # assert MPI.COMM_WORLD.size > 1
    if backend == "adios2":
        backend_args = {"engine": "BP4"}
    else:
        backend_args = None

    if mesh_fname.suffix == ".xdmf":
        with dolfinx.io.XDMFFile(MPI.COMM_WORLD, mesh_fname, "r") as xdmf:
            mesh = xdmf.read_mesh()
    else:
        mesh = io4dolfinx.read_mesh(
            mesh_fname,
            MPI.COMM_WORLD,
            ghost_mode=dolfinx.mesh.GhostMode.shared_facet,
            backend_args=backend_args,
            backend=backend,
        )

    el = basix.ufl.element(
        family,
        mesh.basix_cell(),
        degree,
        basix.LagrangeVariant.gll_warped,
        shape=(mesh.geometry.dim,),
        dtype=mesh.geometry.x.dtype,
    )

    V = dolfinx.fem.functionspace(mesh, el)
    u = dolfinx.fem.Function(V, name=u_name, dtype=u_dtype)
    io4dolfinx.read_function(u_fname, u, time=0.0, backend_args=backend_args, backend=backend)
    MPI.COMM_WORLD.Barrier()

    u_ex = dolfinx.fem.Function(V, name="exact", dtype=u_dtype)
    u_ex.interpolate(f)
    u_ex.x.scatter_forward()
    atol = 10 * np.finfo(u_dtype).resolution
    np.testing.assert_allclose(u.x.array, u_ex.x.array, atol=atol)  # type: ignore


def write_function_vector(
    write_mesh: bool,
    fname: Path,
    family: str,
    degree: int,
    f: Callable[[np.ndarray], np.ndarray],
    dtype: np.dtype,
    name: str,
    dir: Path,
    backend: typing.Literal["adios2", "h5py"],
) -> Path:
    """Convenience function for writing function to file on the original input mesh"""
    from mpi4py import MPI

    import basix.ufl
    import dolfinx

    import io4dolfinx

    assert MPI.COMM_WORLD.size > 1
    with dolfinx.io.XDMFFile(MPI.COMM_WORLD, fname, "r") as xdmf:
        mesh = xdmf.read_mesh()
    el = basix.ufl.element(family, mesh.basix_cell(), degree, dtype=mesh.geometry.x.dtype)
    V = dolfinx.fem.functionspace(mesh, el)
    uh = dolfinx.fem.Function(V, dtype=dtype)
    uh.interpolate(f)
    uh.name = name
    el_hash = (
        io4dolfinx.utils.element_signature(V)
        .replace(" ", "")
        .replace(",", "")
        .replace("(", "")
        .replace(")", "")
        .replace("[", "")
        .replace("]", "")
    )

    file_hash = f"{el_hash}_{np.dtype(dtype).name}"
    filename = dir / f"mesh_{file_hash}.bp"

    if write_mesh:
        io4dolfinx.write_mesh_input_order(filename, mesh, backend=backend)
    io4dolfinx.write_function_on_input_mesh(filename, uh, time=0.0, backend=backend)
    return filename


def read_function_vector(
    mesh_fname: Path,
    u_fname: Path,
    u_name: str,
    family: str,
    degree: int,
    f: Callable[[np.ndarray], np.ndarray],
    u_dtype: np.dtype,
    backend: typing.Literal["adios2", "h5py"],
):
    """
    Convenience function for reading mesh with IPython-parallel and compare to exact solution
    """

    if mesh_fname.suffix == ".xdmf":
        with dolfinx.io.XDMFFile(MPI.COMM_WORLD, mesh_fname, "r") as xdmf:
            mesh = xdmf.read_mesh()
    elif mesh_fname.suffix == ".bp":
        mesh = io4dolfinx.read_mesh(
            mesh_fname,
            MPI.COMM_WORLD,
            ghost_mode=dolfinx.mesh.GhostMode.shared_facet,
            backend=backend,
        )
    el = basix.ufl.element(family, mesh.basix_cell(), degree)

    V = dolfinx.fem.functionspace(mesh, el)
    u = dolfinx.fem.Function(V, name=u_name, dtype=u_dtype)
    io4dolfinx.read_function(u_fname, u, time=0.0, backend=backend)
    MPI.COMM_WORLD.Barrier()

    u_ex = dolfinx.fem.Function(V, name="exact", dtype=u_dtype)
    u_ex.interpolate(f)
    u_ex.x.scatter_forward()
    atol = 10 * np.finfo(u_dtype).resolution
    np.testing.assert_allclose(u.x.array, u_ex.x.array, atol=atol)  # type: ignore


@pytest.mark.skipif(
    os.cpu_count() == 1, reason="Test requires that the system has more than one process"
)
@pytest.mark.skipif(MPI.COMM_WORLD.size > 1, reason="Test uses ipythonparallel for MPI")
@pytest.mark.parametrize("is_complex", [True, False])
@pytest.mark.parametrize("family", ["Lagrange", "DG"])
@pytest.mark.parametrize("degree", [1, 4])
@pytest.mark.parametrize("write_mesh", [True, False])
def test_read_write_P_2D(
    write_mesh, family, degree, is_complex, create_2D_mesh, cluster, get_dtype, tmp_path, backend
):
    fname = create_2D_mesh
    with dolfinx.io.XDMFFile(MPI.COMM_WORLD, fname, "r") as xdmf:
        mesh = xdmf.read_mesh()
    f_dtype = get_dtype(mesh.geometry.x.dtype, is_complex)

    el = basix.ufl.element(
        family,
        mesh.basix_cell(),
        degree,
        basix.LagrangeVariant.gll_warped,
        shape=(mesh.geometry.dim,),
        dtype=mesh.geometry.x.dtype,
    )

    def f(x):
        values = np.empty((2, x.shape[1]), dtype=f_dtype)
        values[0] = np.full(x.shape[1], np.pi) + x[0]
        values[1] = x[0]
        if is_complex:
            values[0] -= 3j * x[1]
            values[1] += 2j * x[0]
        return values

    hash = write_function_original(
        write_mesh, mesh, el, f, f_dtype, "u_original", tmp_path, backend
    )
    if write_mesh:
        mesh_fname = hash
    else:
        mesh_fname = fname
    query = cluster[:].apply_async(
        read_function_original, mesh_fname, hash, "u_original", family, degree, f, f_dtype, backend
    )
    query.wait()
    assert query.successful(), query.error


@pytest.mark.skipif(
    os.cpu_count() == 1, reason="Test requires that the system has more than one process"
)
@pytest.mark.skipif(MPI.COMM_WORLD.size > 1, reason="Test uses ipythonparallel for MPI")
@pytest.mark.parametrize("is_complex", [True, False])
@pytest.mark.parametrize("family", ["Lagrange", "DG"])
@pytest.mark.parametrize("degree", [1, 4])
@pytest.mark.parametrize("write_mesh", [True, False])
def test_read_write_P_3D(
    write_mesh, family, degree, is_complex, create_3D_mesh, cluster, get_dtype, tmp_path, backend
):
    fname = create_3D_mesh
    with dolfinx.io.XDMFFile(MPI.COMM_WORLD, fname, "r") as xdmf:
        mesh = xdmf.read_mesh()
    f_dtype = get_dtype(mesh.geometry.x.dtype, is_complex)
    el = basix.ufl.element(
        family,
        mesh.basix_cell(),
        degree,
        basix.LagrangeVariant.gll_warped,
        shape=(mesh.geometry.dim,),
    )

    def f(x):
        values = np.empty((3, x.shape[1]), dtype=f_dtype)
        values[0] = np.pi + x[0]
        values[1] = x[1] + 2 * x[0]
        values[2] = np.cos(x[2])
        if is_complex:
            values[0] -= np.pi * x[1]
            values[1] += 3j * x[2]
            values[2] += 2j
        return values

    hash = write_function_original(
        write_mesh, mesh, el, f, f_dtype, "u_original", tmp_path, backend=backend
    )
    MPI.COMM_WORLD.Barrier()

    if write_mesh:
        mesh_fname = hash
    else:
        mesh_fname = fname

    query = cluster[:].apply_async(
        read_function_original,
        mesh_fname,
        hash,
        "u_original",
        family,
        degree,
        f,
        f_dtype,
        backend=backend,
    )
    query.wait()
    assert query.successful(), query.error


@pytest.mark.skipif(
    os.cpu_count() == 1, reason="Test requires that the system has more than one process"
)
@pytest.mark.skipif(MPI.COMM_WORLD.size > 1, reason="Test uses ipythonparallel for MPI")
@pytest.mark.parametrize("write_mesh", [True, False])
@pytest.mark.parametrize("is_complex", [True, False])
@pytest.mark.parametrize("family", ["N1curl", "RT"])
@pytest.mark.parametrize("degree", [1, 4])
def test_read_write_2D_vector_simplex(
    write_mesh,
    family,
    degree,
    is_complex,
    create_simplex_mesh_2D,
    cluster,
    get_dtype,
    tmp_path,
    backend,
):
    fname = create_simplex_mesh_2D

    f_dtype = get_dtype(np.float64, is_complex)

    def f(x):
        values = np.empty((2, x.shape[1]), dtype=f_dtype)
        values[0] = np.full(x.shape[1], np.pi) + x[0]
        values[1] = x[1]
        if is_complex:
            values[0] -= np.sin(x[1]) * 2j
            values[1] += 3j
        return values

    query = cluster[:].apply_async(
        write_function_vector,
        write_mesh,
        fname,
        family,
        degree,
        f,
        f_dtype,
        "u_original",
        tmp_path,
        backend,
    )
    query.wait()
    assert query.successful(), query.error
    paths = query.result()
    file_path = paths[0]
    assert all([file_path == path for path in paths])
    if write_mesh:
        mesh_fname = file_path
    else:
        mesh_fname = fname

    read_function_vector(mesh_fname, file_path, "u_original", family, degree, f, f_dtype, backend)


@pytest.mark.skipif(
    os.cpu_count() == 1, reason="Test requires that the system has more than one process"
)
@pytest.mark.skipif(MPI.COMM_WORLD.size > 1, reason="Test uses ipythonparallel for MPI")
@pytest.mark.parametrize("write_mesh", [True, False])
@pytest.mark.parametrize("is_complex", [True, False])
@pytest.mark.parametrize("family", ["N1curl", "RT"])
@pytest.mark.parametrize("degree", [1, 4])
def test_read_write_3D_vector_simplex(
    write_mesh,
    family,
    degree,
    is_complex,
    create_simplex_mesh_3D,
    cluster,
    get_dtype,
    tmp_path,
    backend,
):
    fname = create_simplex_mesh_3D

    f_dtype = get_dtype(np.float64, is_complex)

    def f(x):
        values = np.empty((3, x.shape[1]), dtype=f_dtype)
        values[0] = np.full(x.shape[1], np.pi)
        values[1] = x[1] + 2 * x[0]
        values[2] = np.cos(x[2])
        if is_complex:
            values[0] += 2j * x[2]
            values[1] += 2j * np.cos(x[2])
        return values

    query = cluster[:].apply_async(
        write_function_vector,
        write_mesh,
        fname,
        family,
        degree,
        f,
        f_dtype,
        "u_original",
        tmp_path,
        backend,
    )
    query.wait()
    assert query.successful(), query.error
    paths = query.result()
    file_path = paths[0]
    assert all([file_path == path for path in paths])
    if write_mesh:
        mesh_fname = file_path
    else:
        mesh_fname = fname

    read_function_vector(mesh_fname, file_path, "u_original", family, degree, f, f_dtype, backend)


@pytest.mark.skipif(
    os.cpu_count() == 1, reason="Test requires that the system has more than one process"
)
@pytest.mark.skipif(MPI.COMM_WORLD.size > 1, reason="Test uses ipythonparallel for MPI")
@pytest.mark.parametrize("write_mesh", [True, False])
@pytest.mark.parametrize("is_complex", [True, False])
@pytest.mark.parametrize("family", ["RTCF"])
@pytest.mark.parametrize("degree", [1, 2, 3])
def test_read_write_2D_vector_non_simplex(
    write_mesh,
    family,
    degree,
    is_complex,
    create_non_simplex_mesh_2D,
    cluster,
    get_dtype,
    tmp_path,
    backend,
):
    fname = create_non_simplex_mesh_2D

    f_dtype = get_dtype(np.float64, is_complex)

    def f(x):
        values = np.empty((2, x.shape[1]), dtype=f_dtype)
        values[0] = np.full(x.shape[1], np.pi)
        values[1] = x[1] + 2 * x[0]
        if is_complex:
            values[0] += 2j * x[1]
            values[1] -= np.sin(x[0]) * 9j
        return values

    query = cluster[:].apply_async(
        write_function_vector,
        write_mesh,
        fname,
        family,
        degree,
        f,
        f_dtype,
        "u_original",
        tmp_path,
        backend,
    )
    query.wait()
    assert query.successful(), query.error
    paths = query.result()
    file_path = paths[0]
    assert all([file_path == path for path in paths])
    if write_mesh:
        mesh_fname = file_path
    else:
        mesh_fname = fname

    read_function_vector(mesh_fname, file_path, "u_original", family, degree, f, f_dtype, backend)


@pytest.mark.skipif(
    os.cpu_count() == 1, reason="Test requires that the system has more than one process"
)
@pytest.mark.skipif(MPI.COMM_WORLD.size > 1, reason="Test uses ipythonparallel for MPI")
@pytest.mark.parametrize("write_mesh", [True, False])
@pytest.mark.parametrize("is_complex", [True, False])
@pytest.mark.parametrize("family", ["NCF"])
@pytest.mark.parametrize("degree", [1, 4])
def test_read_write_3D_vector_non_simplex(
    write_mesh,
    family,
    degree,
    is_complex,
    create_non_simplex_mesh_3D,
    cluster,
    get_dtype,
    tmp_path,
    backend,
):
    fname = create_non_simplex_mesh_3D

    f_dtype = get_dtype(np.float64, is_complex)

    def f(x):
        values = np.empty((3, x.shape[1]), dtype=f_dtype)
        values[0] = np.full(x.shape[1], np.pi) + x[0]
        values[1] = np.cos(x[2])
        values[2] = x[0]
        if is_complex:
            values[2] += x[0] * x[1] * 3j
        return values

    query = cluster[:].apply_async(
        write_function_vector,
        write_mesh,
        fname,
        family,
        degree,
        f,
        f_dtype,
        "u_original",
        tmp_path,
        backend,
    )
    query.wait()
    assert query.successful(), query.error
    paths = query.result()
    file_path = paths[0]
    assert all([file_path == path for path in paths])
    if write_mesh:
        mesh_fname = file_path
    else:
        mesh_fname = fname

    read_function_vector(mesh_fname, file_path, "u_original", family, degree, f, f_dtype, backend)



import itertools
from typing import Tuple

from mpi4py import MPI

import basix.ufl
import dolfinx
import numpy as np
import numpy.typing as npt
import pytest

from io4dolfinx.utils import compute_dofmap_pos, unroll_dofmap

write_comm = [MPI.COMM_SELF, MPI.COMM_WORLD]  # Communicators for creating mesh
ghost_mode = [dolfinx.mesh.GhostMode.none, dolfinx.mesh.GhostMode.shared_facet]
two_dimensional_cell_types = [
    dolfinx.mesh.CellType.triangle,
    dolfinx.mesh.CellType.quadrilateral,
]
three_dimensional_cell_types = [dolfinx.mesh.CellType.hexahedron]

two_dim_combinations = itertools.product(two_dimensional_cell_types, write_comm, ghost_mode)
three_dim_combinations = itertools.product(three_dimensional_cell_types, write_comm, ghost_mode)


@pytest.fixture(params=two_dim_combinations, scope="module")
def mesh_2D(request):
    cell_type, write_comm, ghost_mode = request.param
    mesh = dolfinx.mesh.create_unit_square(
        write_comm, 10, 10, cell_type=cell_type, ghost_mode=ghost_mode
    )
    return mesh


@pytest.fixture(params=three_dim_combinations, scope="module")
def mesh_3D(request):
    cell_type, write_comm, ghost_mode = request.param
    M = 5
    mesh = dolfinx.mesh.create_unit_cube(
        write_comm, M, M, M, cell_type=cell_type, ghost_mode=ghost_mode
    )
    return mesh


def compute_positions(
    dofs: npt.NDArray[np.int32],
    dofmap_bs: int,
    num_owned_dofs: int,
    num_owned_cells: int,
) -> Tuple[npt.NDArray[np.int32], npt.NDArray[np.int32]]:
    """
    Support function for test.
    Given a dofmap, compute the local cell and position in the dofmap for each owned dof.
    The last cell (wrt) local index will be the one in the output map
    """
    dof_to_cell_map = np.zeros(num_owned_dofs, dtype=np.int32)
    dof_to_pos_map = np.zeros(num_owned_dofs, dtype=np.int32)
    for c in range(num_owned_cells):
        for i, dof in enumerate(dofs[c]):
            for b in range(dofmap_bs):
                local_dof = dof * dofmap_bs + b
                if local_dof < num_owned_dofs:
                    dof_to_cell_map[local_dof] = c
                    dof_to_pos_map[local_dof] = i * dofmap_bs + b
    return dof_to_cell_map, dof_to_pos_map


@pytest.mark.parametrize("family", ["Lagrange", "DG"])
@pytest.mark.parametrize("degree", [1, 4])
def test_unroll_P(family, degree, mesh_2D):
    V = dolfinx.fem.functionspace(mesh_2D, (family, degree))
    dofmap = V.dofmap

    unrolled_map = unroll_dofmap(dofmap.list, dofmap.bs)

    normal_unroll = np.zeros(
        (dofmap.list.shape[0], dofmap.list.shape[1] * dofmap.bs), dtype=np.int32
    )
    for i, dofs in enumerate(dofmap.list):
        for j, dof in enumerate(dofs):
            for k in range(dofmap.bs):
                normal_unroll[i, j * dofmap.bs + k] = dof * dofmap.bs + k

    np.testing.assert_allclose(unrolled_map, normal_unroll)


@pytest.mark.parametrize("family", ["RTCF"])
@pytest.mark.parametrize("degree", [1, 2, 3])
def test_unroll_RTCF(family, degree, mesh_3D):
    el = basix.ufl.element(family, mesh_3D.basix_cell(), degree)

    V = dolfinx.fem.functionspace(mesh_3D, el)
    dofmap = V.dofmap

    unrolled_map = unroll_dofmap(dofmap.list, dofmap.bs)

    normal_unroll = np.zeros(
        (dofmap.list.shape[0], dofmap.list.shape[1] * dofmap.bs), dtype=np.int32
    )
    for i, dofs in enumerate(dofmap.list):
        for j, dof in enumerate(dofs):
            for k in range(dofmap.bs):
                normal_unroll[i, j * dofmap.bs + k] = dof * dofmap.bs + k

    np.testing.assert_allclose(unrolled_map, normal_unroll)


@pytest.mark.parametrize("family", ["RTCF"])
@pytest.mark.parametrize("degree", [1, 2, 3])
def test_compute_dofmap_pos_RTCF(family, degree, mesh_3D):
    el = basix.ufl.element(family, mesh_3D.basix_cell(), degree)

    V = dolfinx.fem.functionspace(mesh_3D, el)
    local_cells, local_pos = compute_dofmap_pos(V)

    num_cells_local = mesh_3D.topology.index_map(mesh_3D.topology.dim).size_local
    num_dofs_local = V.dofmap.index_map.size_local * V.dofmap.index_map_bs
    reference_cells, reference_pos = compute_positions(
        V.dofmap.list, V.dofmap.bs, num_dofs_local, num_cells_local
    )
    np.testing.assert_allclose(reference_cells, local_cells)
    np.testing.assert_allclose(reference_pos, local_pos)


@pytest.mark.parametrize("family", ["Lagrange", "DG"])
@pytest.mark.parametrize("degree", [1, 4])
def test_compute_dofmap_pos_P(family, degree, mesh_2D):
    el = basix.ufl.element(family, mesh_2D.basix_cell(), degree)

    V = dolfinx.fem.functionspace(mesh_2D, el)
    local_cells, local_pos = compute_dofmap_pos(V)

    num_cells_local = mesh_2D.topology.index_map(mesh_2D.topology.dim).size_local
    num_dofs_local = V.dofmap.index_map.size_local * V.dofmap.index_map_bs
    reference_cells, reference_pos = compute_positions(
        V.dofmap.list, V.dofmap.bs, num_dofs_local, num_cells_local
    )
    np.testing.assert_allclose(reference_cells, local_cells)
    np.testing.assert_allclose(reference_pos, local_pos)


def test_compute_send_sizes():
    np.random.seed(42)
    N = 0
    M = 10
    num_data = 100

    # Set of ranks to recieve data
    dest_ranks = np.arange(N, M, dtype=np.int32)

    # Random data owners
    data_owners = np.random.randint(N, M, num_data).astype(np.int32)

    # Compute the number of data to send to each rank with loops
    out_size = np.zeros(len(dest_ranks), dtype=np.int32)
    for owner in data_owners:
        for j, rank in enumerate(dest_ranks):
            if owner == rank:
                out_size[j] += 1
                break

    process_pos_indicator = data_owners.reshape(-1, 1) == dest_ranks
    vectorized_out_size = np.count_nonzero(process_pos_indicator, axis=0)
    np.testing.assert_allclose(vectorized_out_size, out_size)



from mpi4py import MPI

import numpy as np
import pytest
from packaging.version import parse as _v

import io4dolfinx


@pytest.mark.parametrize("comm", [MPI.COMM_SELF, MPI.COMM_WORLD])
def test_read_write_attributes(comm, backend, tmp_path):
    if backend == "adios2":
        import adios2

        if _v(np.__version__) >= _v("2.0.0") and _v(adios2.__version__) < _v("2.10.2"):
            pytest.skip(reason="Cannot use numpy>=2.0.0 and adios2<2.10.2")

    attributes1 = {
        "a": np.array([1, 2, 3], dtype=np.uint8),
        "b": np.array([4, 5], dtype=np.uint8),
    }
    attributes2 = {
        "c": np.array([6], dtype=np.uint8),
        "d": np.array([7, 8, 9, 10], dtype=np.float64),
    }
    fname = comm.bcast(tmp_path, root=0)
    fname = fname / "attributes"
    suffix = ".bp" if backend == "adios2" else ".h5"
    file = fname.with_suffix(suffix)
    # print(comm.size)

    io4dolfinx.write_attributes(
        comm=comm, filename=file, name="group1", attributes=attributes1, backend=backend
    )
    io4dolfinx.write_attributes(
        comm=comm, filename=file, name="group2", attributes=attributes2, backend=backend
    )
    loaded_attributes1 = io4dolfinx.read_attributes(
        comm=comm, filename=file, name="group1", backend=backend
    )
    loaded_attributes2 = io4dolfinx.read_attributes(
        comm=comm, filename=file, name="group2", backend=backend
    )

    for k, v in loaded_attributes1.items():
        assert np.allclose(v, attributes1[k])
    for k, v in attributes1.items():
        assert np.allclose(v, loaded_attributes1[k])

    for k, v in loaded_attributes2.items():
        assert np.allclose(v, attributes2[k])
    for k, v in attributes2.items():
        assert np.allclose(v, loaded_attributes2[k])



from enum import Enum
from importlib import import_module
from pathlib import Path
from typing import Any, Protocol

from mpi4py import MPI

import dolfinx
import numpy as np
import numpy.typing as npt

from ..structures import ArrayData, FunctionData, MeshData, MeshTagsData, ReadMeshData

__all__ = ["FileMode", "IOBackend", "get_backend"]


class ReadMode(Enum):
    serial = 10  # This means that all data is read in on root rank

    # Total number of data P, num processes = i + 1.
    # All processes reads at least `P // (i+1)` items
    # The first j=P%(i+1) processes reads `P // (i+1) + 1` items
    # ```python
    # def compute_partitioning(P, J):
    #     min_num = P // J
    #     num_per_proc = np.full(J, min_num)
    #     rem = P % J
    #     num_per_proc[:int(rem)] += 1
    #     assert(sum(num_per_proc)) == P
    #     return num_per_proc
    # ```
    parallel = 20


class FileMode(Enum):
    """Filen mode used for opening files."""

    append = 10  #: Append data to file
    write = 20  #: Write data to file
    read = 30  #: Read data from file


# See https://peps.python.org/pep-0544/#modules-as-implementations-of-protocols
class IOBackend(Protocol):
    read_mode: ReadMode

    def get_default_backend_args(self, arguments: dict[str, Any] | None) -> dict[str, Any]:
        """Get default backend arguments given a set of input arguments.

        Args:
            arguments: Input backend arguments

        Returns:
            Updated backend arguments
        """

    def write_attributes(
        self,
        filename: Path | str,
        comm: MPI.Intracomm,
        name: str,
        attributes: dict[str, np.ndarray],
        backend_args: dict[str, Any] | None,
    ):
        """Write attributes to file.

        Args:
            filename: Path to file to write to
            comm: MPI communicator used in storage
            name: Name of the attribute group
            attributes: Dictionary of attributes to write
            backend_args: Arguments to backend
        """

    def read_attributes(
        self,
        filename: Path | str,
        comm: MPI.Intracomm,
        name: str,
        backend_args: dict[str, Any] | None,
    ) -> dict[str, Any]:
        """Read attributes from file.

        Args:
            filename: Path to file to read from
            comm: MPI communicator used in storage
            name: Name of the attribute group
            backend_args: Arguments to backend

        Returns:
            Dictionary of attributes read from file
        """

    def read_timestamps(
        self,
        filename: Path | str,
        comm: MPI.Intracomm,
        function_name: str,
        backend_args: dict[str, Any] | None,
    ) -> npt.NDArray[np.float64 | str]:  # type: ignore[type-var]
        """Read timestamps from file.

        Args:
            filename: Path to file to read from
            comm: MPI communicator used in storage
            function_name: Name of the function to read timestamps for
            backend_args: Arguments to backend

        Returns:
            Numpy array of timestamps read from file
        """

    def write_mesh(
        self,
        filename: Path | str,
        comm: MPI.Intracomm,
        mesh: MeshData,
        backend_args: dict[str, Any] | None,
        mode: FileMode,
        time: float,
    ):
        """
        Write a mesh to file.

        Args:
            comm: MPI communicator used in storage
            mesh: Internal data structure for the mesh data to save to file
            filename: Path to file to write to
            backend_args: Arguments to backend
            mode: File-mode to store the mesh
            time: Time stamp associated with the mesh
        """

    def write_meshtags(
        self,
        filename: str | Path,
        comm: MPI.Intracomm,
        data: MeshTagsData,
        backend_args: dict[str, Any] | None,
    ):
        """Write mesh tags to file.

        Args:
            filename: Path to file to write to
            comm: MPI communicator used in storage
            data: Internal data structure for the mesh tags to save to file
            backend_args: Arguments to backend
        """

    def read_mesh_data(
        self,
        filename: Path | str,
        comm: MPI.Intracomm,
        time: str | float | None,
        read_from_partition: bool,
        backend_args: dict[str, Any] | None,
    ) -> ReadMeshData:
        """Read mesh data from file.

        Args:
            filename: Path to file to read from
            comm: MPI communicator used in storage
            time: Time stamp associated with the mesh to read
            read_from_partition: Whether to read partition information
            backend_args: Arguments to backend

        Returns:
            Internal data structure for the mesh data read from file
        """

    def read_meshtags_data(
        self,
        filename: str | Path,
        comm: MPI.Intracomm,
        name: str,
        backend_args: dict[str, Any] | None,
    ) -> MeshTagsData:
        """Read mesh tags from file.

        Args:
            filename: Path to file to read from
            comm: MPI communicator used in storage
            name: Name of the mesh tags to read
            backend_args: Arguments to backend

        Returns:
            Internal data structure for the mesh tags read from file
        """

    def read_dofmap(
        self,
        filename: str | Path,
        comm: MPI.Intracomm,
        name: str,
        backend_args: dict[str, Any] | None,
    ) -> dolfinx.graph.AdjacencyList:
        """Read the dofmap of a function with a given name.

        Args:
            filename: Path to file to read from
            comm: MPI communicator used in storage
            name: Name of the function to read the dofmap for
            backend_args: Arguments to backend

        Returns:
            Dofmap as an {py:class}`dolfinx.graph.AdjacencyList`
        """

    def read_dofs(
        self,
        filename: str | Path,
        comm: MPI.Intracomm,
        name: str,
        time: float,
        backend_args: dict[str, Any] | None,
    ) -> tuple[npt.NDArray[np.float32 | np.float64 | np.complex64 | np.complex128], int]:
        """Read the dofs (values) of a function with a given name from a given timestep.

        Args:
            filename: Path to file to read from
            comm: MPI communicator used in storage
            name: Name of the function to read the dofs for
            time: Time stamp associated with the function to read
            backend_args: Arguments to backend

        Returns:
            Contiguous sequence of degrees of freedom (with respect to input data)
            and the global starting point on the process.
            Process 0 has [0, M), process 1 [M, N), process 2 [N, O) etc.
        """

    def read_cell_perms(
        self, comm: MPI.Intracomm, filename: Path | str, backend_args: dict[str, Any] | None
    ) -> npt.NDArray[np.uint32]:
        """
        Read cell permutation from file with given communicator,
        Split in continuous chunks based on number of cells in the input data.

        Args:
            comm: MPI communicator used in storage
            filename: Path to file to read from
            backend_args: Arguments to backend

        Returns:
            Contiguous sequence of permutations (with respect to input data)
            Process 0 has [0, M), process 1 [M, N), process 2 [N, O) etc.
        """

    def write_function(
        self,
        filename: Path,
        comm: MPI.Intracomm,
        u: FunctionData,
        time: float,
        mode: FileMode,
        backend_args: dict[str, Any] | None,
    ):
        """Write a function to file.

        Args:
            comm: MPI communicator used in storage
            u: Internal data structure for the function data to save to file
            filename: Path to file to write to
            time: Time stamp associated with function
            mode: File-mode to store the function
            backend_args: Arguments to backend
        """

    def read_legacy_mesh(
        self, filename: Path | str, comm: MPI.Intracomm, group: str
    ) -> tuple[npt.NDArray[np.int64], npt.NDArray[np.floating], str | None]:
        """Read in the mesh topology, geometry and (optionally) cell type from a
        legacy DOLFIN HDF5-file.

        Args:
            filename: Path to file to read from
            comm: MPI communicator used in storage
            group: Group in HDF5 file where mesh is stored

        Returns:
            Tuple containing:
                - Topology as a (num_cells, num_vertices_per_cell) array of global vertex indices
                - Geometry as a (num_vertices, geometric_dimension) array of vertex coordinates
                - Cell type as a string (e.g. "tetrahedron") or None if not found
        """

    def snapshot_checkpoint(
        self,
        filename: Path | str,
        mode: FileMode,
        u: dolfinx.fem.Function,
        backend_args: dict[str, Any] | None,
    ):
        """Create a snapshot checkpoint of a dolfinx function.

        Args:
            filename: Path to file to read from
            mode: File-mode to store the function
            u: dolfinx function to create a snapshot checkpoint for
            backend_args: Arguments to backend
        """

    def read_hdf5_array(
        self,
        comm: MPI.Intracomm,
        filename: Path | str,
        group: str,
        backend_args: dict[str, Any] | None,
    ) -> tuple[np.ndarray, int]:
        """Read an array from an HDF5 file.

        Args:
            comm: MPI communicator used in storage
            filename: Path to file to read from
            group: Group in HDF5 file where array is stored
            backend_args: Arguments to backend

        Returns:
            Tuple containing:
                - Numpy array read from file
                - Global starting point on the process.
                  Process 0 has [0, M), process 1 [M, N), process 2 [N, O) etc.
        """

    def read_point_data(
        self,
        filename: Path | str,
        name: str,
        comm: MPI.Intracomm,
        time: str | float | None,
        backend_args: dict[str, Any] | None,
    ) -> tuple[np.ndarray, int]:
        """Read data from the nodes of a mesh.

        Args:
            filename: Path to file
            name: Name of point data
            comm: Communicator to launch IO on.
            time: The time stamp
            backend_args: The backend arguments
        Returns:
            Data local to process (contiguous, no mpi comm) and local start range
        """
        ...

    def read_function_names(
        self, filename: Path | str, comm: MPI.Intracomm, backend_args: dict[str, Any] | None
    ) -> list[str]:
        """Read all function names from a file.

        Args:
            filename: Path to file
            comm: MPI communicator to launch IO on.
            backend_args: Arguments to backend

        Returns:
            A list of function names.
        """
        ...

    def read_cell_data(
        self,
        filename: Path | str,
        name: str,
        comm: MPI.Intracomm,
        time: str | float | None,
        backend_args: dict[str, Any] | None,
    ) -> tuple[npt.NDArray[np.int64], np.ndarray]:
        """Read data from the cells of a mesh.

        Args:
            filename: Path to file
            name: Name of point data
            comm: Communicator to launch IO on.
            time: The time stamp
            backend_args: The backend arguments
        Returns:
            A tuple (topology, dofs) where topology contains the
            vertex indices of the cells, dofs the degrees of
            freedom within that cell.
        """
        ...

    def write_data(
        self,
        filename: Path | str,
        array_data: ArrayData,
        comm: MPI.Intracomm,
        time: str | float | None,
        mode: FileMode,
        backend_args: dict[str, Any] | None,
    ):
        """Write a 2D-array to file.


        Args:
            filename: Path to file
            array_data: Data to write to file.
            comm: The MPI communicator to open the writer with.
            time: The time stamp
            mode: Append or write
            backend_args: The backend arguments
        """
        ...


def get_backend(backend: str) -> IOBackend:
    """Get backend class from backend name.

    Args:
        backend: Name of the backend to get

    Returns:
        Backend class
    """
    if backend == "h5py":
        from .h5py import backend as H5PYInterface

        return H5PYInterface
    elif backend == "adios2":
        from .adios2 import backend as ADIOS2Interface

        return ADIOS2Interface
    elif backend == "pyvista":
        from .pyvista import backend as PYVISTAInterface

        return PYVISTAInterface
    elif backend == "xdmf":
        from .xdmf import backend as XDMFInterface

        return XDMFInterface
    elif backend == "vtkhdf":
        from .vtkhdf import backend as VTKDHFInterface

        return VTKDHFInterface
    elif backend == "exodus":
        from .exodus import backend as EXODUSInterface

        return EXODUSInterface
    else:
        return import_module(backend)



from mpi4py import MPI

import dolfinx
import numpy as np
import pytest
import ufl

import io4dolfinx

pyvista = pytest.importorskip("pyvista")


def test_read_mesh_and_cell_data(tmp_path):
    tmp_path = MPI.COMM_WORLD.bcast(tmp_path, root=0)
    filename = tmp_path / "grid.vtu"
    grid = pyvista.examples.load_hexbeam()
    if MPI.COMM_WORLD.rank == 0:
        grid.save(filename)
    MPI.COMM_WORLD.barrier()

    mesh = io4dolfinx.read_mesh(filename, MPI.COMM_WORLD, backend="pyvista")

    vol = dolfinx.fem.form(1 * ufl.dx(domain=mesh))
    surf = dolfinx.fem.form(1 * ufl.ds(domain=mesh))

    vol_ref = 5 * 1 * 1
    surf_ref = 5 * 4 + 2

    vol_glob = mesh.comm.allreduce(dolfinx.fem.assemble_scalar(vol), op=MPI.SUM)
    surf_glob = mesh.comm.allreduce(dolfinx.fem.assemble_scalar(surf), op=MPI.SUM)
    assert np.isclose(vol_glob, vol_ref)
    assert np.isclose(surf_glob, surf_ref)

    names = io4dolfinx.read_function_names(filename, MPI.COMM_WORLD, backend="pyvista")
    for name in grid.cell_data.keys():
        assert name in names
    for name in grid.point_data.keys():
        assert name in names

    for name in names:
        if name in grid.cell_data.keys():
            cd = io4dolfinx.read_cell_data(filename, name, mesh, backend="pyvista")
            oci = mesh.topology.original_cell_index
            np.testing.assert_allclose(cd.x.array[:], grid.cell_data[name][oci])
        elif name in grid.point_data.keys():
            pd = io4dolfinx.read_point_data(filename, name, mesh, backend="pyvista")
            igi = mesh.geometry.input_global_indices
            np.testing.assert_allclose(pd.x.array[:], grid.point_data[name][igi])
        else:
            raise RuntimeError(f"Could not find {name} in grid")



# Copyright (C) 2024 Jørgen Schartum Dokken
#
# This file is part of io4dolfinx
#
# SPDX-License-Identifier:    MIT

from pathlib import Path
from typing import Any

import dolfinx

from .backends import FileMode, get_backend

__all__ = [
    "snapshot_checkpoint",
]


def snapshot_checkpoint(
    uh: dolfinx.fem.Function,
    file: Path,
    mode: FileMode,
    backend_args: dict[str, Any] | None = None,
    backend: str = "adios2",
):
    """Read or write a snapshot checkpoint

    This checkpoint is only meant to be used on the same mesh during the same simulation.

    :param uh: The function to write data from or read to
    :param file: The file to write to or read from
    :param mode: Either read or write
    """

    backend_cls = get_backend(backend)
    default_args = backend_cls.get_default_backend_args(backend_args)
    if mode not in [FileMode.write, FileMode.read]:
        raise ValueError(f"Got invalid mode {mode}")
    backend_cls.snapshot_checkpoint(file, mode, uh, default_args)
